@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2018-2021  Bruno Chabrier
 #
 # This file is part of woob.
@@ -17,8 +15,6 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
 # start with:
 # set PYTHONPATH=D:\Dropbox\Projets\boomoney
 # D:\Dropbox\Projets\boomoney\scripts\bin\woob.exe money -N
@@ -39,9 +35,14 @@ from asyncio.subprocess import PIPE
 from asyncio import create_subprocess_exec
 
 import shutil
-from colorama import init, Fore, Style
 
-from woob.tools.compat import unicode
+try:
+    from colorama import init as init_colorama, Fore, Style
+except ModuleNotFoundError:
+    HAVE_COLORAMA = False
+else:
+    HAVE_COLORAMA = True
+
 from woob.capabilities.bank import AccountType
 from woob.applications.bank import Appbank
 from woob.applications.bank.bank import OfxFormatter
@@ -291,7 +292,7 @@ class HistoryThreadAsAProcess(HistoryThread):
 class AppMoney(Appbank):
     APPNAME = 'money'
     OLD_APPNAME = 'boomoney'
-    VERSION = '3.0'
+    VERSION = '3.3.1'
     COPYRIGHT = 'Copyright(C) 2018-YEAR Bruno Chabrier'
     DESCRIPTION = "Console application that imports bank accounts into Microsoft Money"
     SHORT_DESCRIPTION = "import bank accounts into Microsoft Money"
@@ -620,7 +621,7 @@ class AppMoney(Appbank):
 
                 # MSMoney expects CHECKNUM instead of NAME for CHECK transactions
                 if "TRNTYPE" in field and field["TRNTYPE"] == "CHECK":
-                    if "NAME" in field and unicode(field["NAME"]).isnumeric():
+                    if "NAME" in field and str(field["NAME"]).isnumeric():
                         field["CHECKNUM"] = field["NAME"]
                         del field["NAME"]
                         fields = fields.replace(' NAME ', ' CHECKNUM ')
@@ -800,7 +801,11 @@ class AppMoney(Appbank):
 
     def main(self, argv):
 
-        init()
+        if not HAVE_COLORAMA:
+            print('Error: Please install the "colorama" python package', file=sys.stderr)
+            return 1
+
+        init_colorama()
 
         self.load_config()
 
