@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2011  Christophe Benz
 #
 # This file is part of woob.
@@ -17,13 +15,12 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
-import os
 import datetime
 import hashlib
-
+import shlex
+import subprocess
 from tempfile import NamedTemporaryFile
+
 from lxml import etree
 
 from woob.core import CallErrors
@@ -42,56 +39,55 @@ __all__ = ['AppMsg']
 class AtomFormatter(IFormatter):
     MANDATORY_FIELDS = ('title', 'date', 'sender', 'content')
 
+    def _format_date(self, dt):
+        return dt.strftime("%Y-%m-%dT%H:%M:%SZ")
+
     def start_format(self, **kwargs):
-        self.output(u'<?xml version="1.0" encoding="utf-8"?><feed xmlns="http://www.w3.org/2005/Atom"')
-        self.output(u'xmlns:dc="http://purl.org/dc/elements/1.1/">\n')
-        self.output(u'<title type="text">Atom feed by Woob</title>')  # TODO : get backend name
-        self.output(u'<updated>%s</updated>' % datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ"))
-        m = hashlib.md5()
-        m.update(datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ"))
-        self.output(u'<id>urn:md5:%s</id>' % m.hexdigest())
+        gen_time = datetime.datetime.utcnow()
+
+        self.output('<?xml version="1.0" encoding="utf-8"?>')
+        self.output('<feed xmlns="http://www.w3.org/2005/Atom" xmlns:dc="http://purl.org/dc/elements/1.1/">')
+        self.output('  <title type="text">Atom feed by Woob</title>')  # TODO : get backend name
+        self.output('  <updated>%s</updated>' % self._format_date(gen_time))
+
+        m = hashlib.new("md5")  # nosec
+        m.update(self._format_date(gen_time).encode("ascii"))
+        self.output("  <id>urn:md5:%s</id>" % m.hexdigest())
 
     def format_obj(self, obj, alias):
-        elem = etree.Element('entry')
+        elem = etree.Element("entry")
 
-        title = etree.Element('title')
+        title = etree.SubElement(elem, "title")
         title.text = obj.title
-        elem.append(title)
 
-        id = etree.Element('id')
-        m = hashlib.md5()
-        m.update(obj.content.encode('utf8', 'ascii'))
+        id = etree.SubElement(elem, "id")
+        m = hashlib.new("md5")  # nosec
+        m.update(obj.content.encode("utf8"))
         id.text = "urn:md5:%s" % m.hexdigest()
-        elem.append(id)
 
-        link = etree.Element('link')
-        link.attrib["href"] = obj.thread.id
+        link = etree.SubElement(elem, "link")
+        link.attrib["href"] = obj.url
         link.attrib["title"] = obj.title
         link.attrib["type"] = "text/html"
-        elem.append(link)
 
-        author = etree.Element('author')
-        name = etree.Element('name')
+        author = etree.SubElement(elem, "author")
+        name = etree.SubElement(author, "name")
         if obj.sender:
             name.text = obj.sender
         else:
             name.text = obj.backend
-        author.append(name)
-        elem.append(author)
 
-        date = etree.Element('updated')
-        date.text = obj.date.strftime("%Y-%m-%dT%H:%M:%SZ")
-        elem.append(date)
+        date = etree.SubElement(elem, "updated")
+        date.text = self._format_date(obj.date)
 
-        content = etree.Element('content')
+        content = etree.SubElement(elem, 'content')
         content.text = obj.content
         content.attrib["type"] = "html"
-        elem.append(content)
 
-        return etree.tostring(elem, pretty_print=True)
+        return etree.tostring(elem, pretty_print=True, encoding="unicode")
 
     def flush(self):
-        self.output(u'</feed>')
+        self.output("</feed>")
 
 
 class XHtmlFormatter(IFormatter):
@@ -114,14 +110,14 @@ class MessageFormatter(IFormatter):
     MANDATORY_FIELDS = ('title', 'date', 'sender', 'signature', 'content')
 
     def format_obj(self, obj, alias):
-        result = u'%sTitle:%s %s\n' % (self.BOLD,
+        result = '%sTitle:%s %s\n' % (self.BOLD,
                                        self.NC, obj.title)
-        result += u'%sDate:%s %s\n' % (self.BOLD,
+        result += '%sDate:%s %s\n' % (self.BOLD,
                                        self.NC, obj.date.strftime('%Y-%m-%d %H:%M'))
-        result += u'%sFrom:%s %s\n' % (self.BOLD,
+        result += '%sFrom:%s %s\n' % (self.BOLD,
                                        self.NC, obj.sender)
         if hasattr(obj, 'receivers') and obj.receivers:
-            result += u'%sTo:%s %s\n' % (self.BOLD,
+            result += '%sTo:%s %s\n' % (self.BOLD,
                                          self.NC,
                                          ', '.join(obj.receivers))
 
@@ -159,16 +155,16 @@ class MessagesListFormatter(IFormatter):
     def format_dict_thread(self, obj, alias):
         self.count += 1
         if self.interactive:
-            result = u'%s* (%d) %s (%s)%s' % (self.BOLD,
+            result = '%s* (%d) %s (%s)%s' % (self.BOLD,
                                               self.count,
                                               obj.title, obj.backend,
                                               self.NC)
         else:
-            result = u'%s* (%s) %s%s' % (self.BOLD, obj.id,
+            result = '%s* (%s) %s%s' % (self.BOLD, obj.id,
                                          obj.title,
                                          self.NC)
         if obj.date:
-            result += u'\n             %s' % obj.date
+            result += '\n             %s' % obj.date
         return result
 
     def format_dict_messages(self, obj, alias):
@@ -182,7 +178,7 @@ class MessagesListFormatter(IFormatter):
 
     def format_message(self, backend, message, depth=0):
         if not message:
-            return u''
+            return ''
         self.count += 1
 
         flags = '['
@@ -199,7 +195,7 @@ class MessagesListFormatter(IFormatter):
         flags += ']'
 
         if self.interactive:
-            result = u'%s%s* (%d)%s %s <%s> %s (%s)\n' % (depth * '  ',
+            result = '%s%s* (%d)%s %s <%s> %s (%s)\n' % (depth * '  ',
                                                           self.BOLD,
                                                           self.count,
                                                           self.NC,
@@ -208,7 +204,7 @@ class MessagesListFormatter(IFormatter):
                                                           message.title,
                                                           backend)
         else:
-            result = u'%s%s* (%s.%s@%s)%s %s <%s> %s\n' % (depth * '  ',
+            result = '%s%s* (%s.%s@%s)%s %s <%s> %s\n' % (depth * '  ',
                                                            self.BOLD,
                                                            message.thread.id,
                                                            message.id,
@@ -236,7 +232,7 @@ class ProfileFormatter(IFormatter):
 class AppMsg(ReplApplication):
     APPNAME = 'msg'
     OLD_APPNAME = 'boobmsg'
-    VERSION = '3.0'
+    VERSION = '3.3.1'
     COPYRIGHT = 'Copyright(C) 2010-YEAR Christophe Benz'
     DESCRIPTION = "Console application allowing to send messages on various websites and " \
                   "to display message threads and contents."
@@ -268,7 +264,7 @@ class AppMsg(ReplApplication):
 
     def main(self, argv):
         self.load_config()
-        return ReplApplication.main(self, argv)
+        return super().main(argv)
 
     def do_status(self, line):
         """
@@ -314,14 +310,14 @@ class AppMsg(ReplApplication):
             text = self.acquire_input()
 
         if not self.options.accept_empty and not text.strip():
-            self.logger.warning(u'The message body is empty, use option --accept_empty to send empty messages')
+            self.logger.warning('The message body is empty, use option --accept_empty to send empty messages')
             return
 
         for receiver in receivers.strip().split(','):
             receiver, backend_name = self.parse_id(receiver.strip(),
                                                    unique_backend=True)
             if not backend_name and len(self.enabled_backends) > 1:
-                self.logger.warning(u'No backend specified for receiver "%s": message will be sent with all the '
+                self.logger.warning('No backend specified for receiver "%s": message will be sent with all the '
                     'enabled backends (%s)' % (receiver,
                     ','.join(backend.name for backend in self.enabled_backends)))
 
@@ -384,11 +380,17 @@ class AppMsg(ReplApplication):
             if not thread:
                 continue
             if len(arg) > 0:
+                if not thread.root:
+                    thread, = self.do("fillobj", thread, ("root",), backends=thread.backend)
+                if thread.root:
+                    thread.root, = self.do("fillobj", thread.root, ("children",), backends=thread.backend)
+
                 for m in thread.iter_all_messages():
                     if not m.backend:
                         m.backend = thread.backend
                     self.messages.append(m)
             else:
+                thread, = self.do("fillobj", thread, ("title", "date"), backends=thread.backend)
                 self.threads.append(thread)
             self.format(thread)
 
@@ -469,6 +471,8 @@ class AppMsg(ReplApplication):
                             thread, = self.do('fillobj', thread, ('root',), backends=thread.backend)
                         message = thread.root
         if not empty(message):
+            message, = self.do("fillobj", message, backends=message.backend)
+
             self.start_format()
             self.format(message)
             self.woob.do('set_message_read', message, backends=message.backend)
@@ -492,7 +496,7 @@ class AppMsg(ReplApplication):
                 found = 1
 
         if not found:
-            self.logger.error(u'Profile not found')
+            self.logger.error('Profile not found')
 
     def do_photos(self, id):
         """
@@ -505,6 +509,10 @@ class AppMsg(ReplApplication):
             print("Configuration error: photo_viewer is undefined", file=self.stderr)
             return
 
+        photo_cmd = shlex.split(photo_cmd)
+        if photo_cmd[-1] == '%s':
+            del photo_cmd
+
         _id, backend_name = self.parse_id(id, unique_backend=True)
 
         found = 0
@@ -521,8 +529,9 @@ class AppMsg(ReplApplication):
                     photo = self.woob[contact.backend].fillobj(photo, 'data')
                     f.write(photo.data)
                     tmp_files.append(f)
-                os.system(photo_cmd % ' '.join([file.name for file in tmp_files]))
+
+                subprocess.call(photo_cmd + [file.name for file in tmp_files])
                 found = 1
 
         if not found:
-            self.logger.error(u'Profile not found')
+            self.logger.error('Profile not found')
