@@ -1,6 +1,4 @@
-# -*- coding: utf-8 -*-
-
-# Copyright(C) 2012-2014 Laurent Bachelier
+# Copyright(C) 2012-2021 Romain Bignon
 #
 # This file is part of woob.
 #
@@ -17,59 +15,49 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import absolute_import, print_function
-
 from collections import OrderedDict
 from functools import wraps
+import importlib
 import re
-import pickle
 import base64
-import io
 from hashlib import sha256
 import zlib
-from functools import reduce
 try:
     from requests.packages import urllib3
 except ImportError:
     import urllib3
 import os
 import sys
-from copy import deepcopy
+from copy import copy, deepcopy
 import inspect
 from datetime import datetime, timedelta
-from dateutil import parser
+from dateutil import parser, tz
 from threading import Lock
+from urllib.parse import urlparse, urljoin, urlencode, parse_qsl
 from uuid import uuid4
 
-try:
-    import requests
-    if int(requests.__version__.split('.')[0]) < 2:
-        raise ImportError()
-except ImportError:
-    raise ImportError('Please install python3-requests >= 2.0')
+import requests
 
 from woob.exceptions import (
-    BrowserHTTPSDowngrade, ModuleInstallError, BrowserRedirect, BrowserIncorrectPassword,
-    NeedInteractiveFor2FA, BrowserInteraction,
+    BrowserHTTPSDowngrade, BrowserRedirect, BrowserIncorrectPassword,
+    BrowserUnavailable,
 )
 
+from woob.tools.date import now_as_utc
 from woob.tools.log import getLogger
-from woob.tools.compat import basestring, unicode, urlparse, urljoin, urlencode, parse_qsl
-from woob.tools.misc import to_unicode
 from woob.tools.json import json
-from woob.tools.value import Value
-from woob import __version__
 
 from .adapters import HTTPAdapter
 from .cookies import WoobCookieJar
 from .exceptions import HTTPNotFound, ClientError, ServerError
+from .har import HARManager
 from .sessions import FuturesSession
 from .profiles import Firefox
 from .pages import NextPage
 from .url import URL, normalize_url
 
 
-class Browser(object):
+class Browser:
     """
     Simple browser class.
     Act like a browser, and don't try to do too much.
@@ -111,6 +99,8 @@ class Browser(object):
     Controls the behavior of get_referrer.
     """
 
+    HTTP_ADAPTER_CLASS = HTTPAdapter
+
     COOKIE_POLICY = None
     """
     Default CookieJar policy.
@@ -137,12 +127,16 @@ class Browser(object):
         return object.__new__(cls)
 
     def __init__(self, logger=None, proxy=None, responses_dirname=None, woob=None, proxy_headers=None, weboob=None):
-        self.logger = getLogger('browser', logger)
+        if logger:
+            self.logger = getLogger("browser", logger)
+        else:
+            self.logger = getLogger("woob.browser")
+
         self.responses_dirname = responses_dirname
         self.responses_count = 0
         self.responses_lock = Lock()
 
-        if isinstance(self.VERIFY, basestring):
+        if isinstance(self.VERIFY, str):
             self.VERIFY = self.asset(self.VERIFY)
 
         self.PROXIES = proxy
@@ -150,7 +144,10 @@ class Browser(object):
         self._setup_session(self.PROFILE)
         self.url = None
         self.response = None
-        self.har_bundle = None
+        self.har_manager = None
+
+        if self.responses_dirname is not None:
+            self.har_manager = HARManager(self.responses_dirname, self.logger)
 
     def deinit(self):
         self.session.close()
@@ -162,6 +159,7 @@ class Browser(object):
         if self.responses_dirname is None:
             import tempfile
             self.responses_dirname = tempfile.mkdtemp(prefix='woob_session_')
+            self.har_manager = HARManager(self.responses_dirname, self.logger)
             print('Debug data will be saved in this directory: %s' % self.responses_dirname, file=sys.stderr)
         elif not os.path.isdir(self.responses_dirname):
             os.makedirs(self.responses_dirname)
@@ -197,7 +195,8 @@ class Browser(object):
                 for key, value in request.headers.items():
                     f.write('%s: %s\n' % (key, value))
                 if request.body is not None:  # separate '' from None
-                    f.write('\n\n\n%s' % request.body)
+                    body = request.body if isinstance(request.body, str) else request.body.decode()
+                    f.write('\n\n\n%s' % body)
             with open(response_filepath + '-response.txt', 'w') as f:
                 if hasattr(response.elapsed, 'total_seconds'):
                     f.write('Time: %3.3fs\n' % response.elapsed.total_seconds())
@@ -213,163 +212,29 @@ class Browser(object):
                 f.write('# %d %s %s\n' % (response.status_code, response.reason, response.headers.get('Content-Type', '')))
                 f.write('%s\t%s\n' % (response.url, filename))
 
-        request = response.request
+        self.har_manager.save_response(slug, response)
 
-        if not self.har_bundle:
-            self.har_bundle = {
-                'log': {
-                    'version': '1.2',
-                    'creator': {
-                        'name': 'woob',
-                        'version': __version__,
-                    },
-                    'browser': {
-                        'name': 'woob',
-                        'version': __version__,
-                    },
-                    # there are no pages, but we need that to please firefox
-                    'pages': [{
-                        'id': 'fake_page',
-                        'pageTimings': {},
-                        # and chromium wants some of it too
-                        'startedDateTime': (datetime.now() - response.elapsed).isoformat(),
-                    }],
-                    # don't put additional data after this list, to have a fixed-size suffix after it
-                    # so we can add more entries without rewriting the whole file.
-                    'entries': [],
-                },
-            }
-
-        har_entry = {
-            '$anchor': slug,
-            'startedDateTime': (datetime.now() - response.elapsed).isoformat(),
-            'pageref': 'fake_page',
-            'time': int(response.elapsed.total_seconds() * 1000),
-            'request': {
-                'method': request.method,
-                'url': request.url,
-                'httpVersion': 'HTTP/%.1f' % (response.raw.version / 10.),
-                'headers': [
-                    {
-                        'name': k,
-                        'value': v,
-                    }
-                    for k, v in request.headers.items()
-                ],
-                'queryString': [
-                    {
-                        'name': key,
-                        'value': value,
-                    }
-                    for key, value in parse_qsl(
-                        urlparse(request.url).query,
-                        keep_blank_values=True,
-                    )
-                ],
-                'cookies': [
-                    {
-                        'name': k,
-                        'value': v,
-                    }
-                    for k, v in request._cookies.items()
-                ],
-                # for chromium
-                'bodySize': -1,
-                'headersSize': -1,
-            },
-            'response': {
-                'status': response.status_code,
-                'statusText': to_unicode(response.reason),
-                'httpVersion': 'HTTP/%.1f' % (response.raw.version / 10.),
-                'headers': [
-                    {
-                        'name': k,
-                        'value': v,
-                    }
-                    for k, v in response.headers.items()
-                ],
-                'content': {
-                    'mimeType': response.headers.get('Content-Type', ''),
-                    'size': len(response.content),
-                    # systematically use base64 to avoid more content alteration
-                    # than there already is...
-                    'encoding': "base64",
-                    'text': base64.b64encode(response.content).decode('ascii'),
-                },
-                'cookies': [
-                    {
-                        'name': k,
-                        'value': v,
-                    }
-                    for k, v in response.cookies.items()
-                ],
-                'redirectURL': response.headers.get('location', ''),
-                # for chromium
-                'bodySize': -1,
-                'headersSize': -1,
-            },
-            'timings': {  # please chromium
-                'send': -1,
-                'wait': -1,
-                'receive': -1,
-            },
-            'cache': {},
-        }
-        if request.body is not None:
-            har_entry['request']['postData'] = {
-                'mimeType': request.headers.get('Content-Type', ''),
-                'params': [],
-            }
-            if isinstance(request.body, str):
-                har_entry['request']['postData']['text'] = request.body
-            else:
-                # HAR format has no proper way to encode posted binary data!
-                har_entry['request']['postData']['text'] = request.body.decode('latin-1')
-                # add a non-standard key to indicate how should "text" be decoded.
-                har_entry['request']['postData']['x-binary'] = True
-
-            if request.headers.get('Content-Type') == 'application/x-www-form-urlencoded':
-                har_entry['request']['postData']['params'] = [
-                    {
-                        "name": key,
-                        "value": value,
-                    } for key, value in parse_qsl(request.body)
-                ]
-
-        with self.responses_lock:
-            self.har_bundle['log']['entries'].append(har_entry)
-
-            har_path = os.path.join(self.responses_dirname, 'bundle.har')
-            if not os.path.isfile(har_path):
-                with open(har_path, 'w') as fd:
-                    json.dump(self.har_bundle, fd, separators=(',', ':'))
-            else:
-                # hack to avoid rewriting the whole file: entries are last in the JSON file
-                # we need to seek at the right place and write the new entry.
-                # this will unfortunately overwrite closings.
-                suffix = "]}}"
-                with open(har_path, 'r+') as fd:
-                    # can't seek with a negative value...
-                    fd.seek(0, io.SEEK_END)
-                    after_entry_pos = fd.tell() - len(suffix)
-                    fd.seek(after_entry_pos)
-
-                    if fd.read(len(suffix)) != suffix:
-                        self.logger.warning('HAR file does not end with the expected pattern')
-                    else:
-                        fd.seek(after_entry_pos)
-                        fd.write(',')  # there should have been at least one entry
-                        json.dump(har_entry, fd, separators=(',', ':'))
-                        fd.write(suffix)
-
-        msg = u'Response saved to %s' % response_filepath
+        msg = 'Response saved to %s' % response_filepath
         if warning:
             self.logger.warning(msg)
         else:
             self.logger.info(msg)
 
+    def _save_request_to_har(self, request):
+        # called when we don't have any response object
+        if not os.path.isdir(self.responses_dirname):
+            os.makedirs(self.responses_dirname)
+
+        slug = uuid4().hex
+        time = self.TIMEOUT * 1000  # because TIMEOUT is in seconds, and we want milliseconds
+        self.har_manager.save_request_only(slug, request, time)
+        self.logger.warning('Request saved to %s' % slug)
+
     def _create_session(self):
-        return FuturesSession(max_workers=self.MAX_WORKERS, max_retries=self.MAX_RETRIES)
+        return FuturesSession(
+            max_workers=self.MAX_WORKERS, max_retries=self.MAX_RETRIES,
+            adapter_class=self.HTTP_ADAPTER_CLASS,
+        )
 
     def _setup_session(self, profile):
         """
@@ -395,8 +260,8 @@ class Browser(object):
         if self.MAX_WORKERS > requests.adapters.DEFAULT_POOLSIZE:
             adapter_kwargs.update(pool_connections=self.MAX_WORKERS,
                                   pool_maxsize=self.MAX_WORKERS)
-        session.mount('https://', HTTPAdapter(**adapter_kwargs))
-        session.mount('http://', HTTPAdapter(**adapter_kwargs))
+        session.mount('https://', self.HTTP_ADAPTER_CLASS(**adapter_kwargs))
+        session.mount('http://', self.HTTP_ADAPTER_CLASS(**adapter_kwargs))
 
         if self.TIMEOUT:
             session.timeout = self.TIMEOUT
@@ -431,7 +296,7 @@ class Browser(object):
         self.url = self.response.url
         return response
 
-    def open(self, url, referrer=None,
+    def open(self, url, *, referrer=None,
                    allow_redirects=True,
                    stream=None,
                    timeout=None,
@@ -487,18 +352,13 @@ class Browser(object):
 
         :rtype: :class:`requests.Response`
         """
-        if 'async' in kwargs:
-            import warnings
-            warnings.warn('Please use is_async instead of async.', DeprecationWarning)
-            is_async = kwargs['async']
-            del kwargs['async']
 
-        if isinstance(url, basestring):
+        if isinstance(url, str):
             url = normalize_url(url)
         elif isinstance(url, requests.Request):
             url.url = normalize_url(url.url)
 
-        req = self.build_request(url, referrer, data_encoding=data_encoding, **kwargs)
+        req = self.build_request(url, referrer=referrer, data_encoding=data_encoding, **kwargs)
         preq = self.prepare_request(req)
 
         if hasattr(preq, '_cookies'):
@@ -528,15 +388,42 @@ class Browser(object):
             return callback(response)
 
         # call python3-requests
-        response = self.session.send(preq,
-                                     allow_redirects=allow_redirects,
-                                     stream=stream,
-                                     timeout=timeout,
-                                     verify=verify,
-                                     cert=cert,
-                                     proxies=proxies,
-                                     callback=inner_callback,
-                                     is_async=is_async)
+        try:
+            response = self.session.send(preq,
+                                         allow_redirects=allow_redirects,
+                                         stream=stream,
+                                         timeout=timeout,
+                                         verify=verify,
+                                         cert=cert,
+                                         proxies=proxies,
+                                         callback=inner_callback,
+                                         is_async=is_async)
+        except Exception as error:
+            # response in these kind of exception are already stored in HAR
+            # skip them to not store response twice
+            already_handled_exception = any(
+                isinstance(error, exc) for exc in (HTTPNotFound, ClientError, ServerError)
+            )
+            if not already_handled_exception and self.responses_dirname is not None:
+                # when timeout or any kind of exception occur
+                # we don't have any response given by python-requests
+                # but we still need to store request to HAR
+                if isinstance(error, requests.exceptions.RequestException):
+                    if error.response is not None:
+                        slug = uuid4().hex
+                        self.har_manager.save_response(slug, error.response)
+                    elif error.request:
+                        # some exceptions raised by python-requests don't even provide request
+                        # we prefer use error.request because python3-requests may modify request
+                        # by calling add_headers() to it, but it won't happens if
+                        # requests.adapters.HTTPAdapter.add_headers() is not overwritten
+                        self._save_request_to_har(error.request)
+                    else:
+                        # use preq request if the exception doesn't provide the request
+                        self._save_request_to_har(preq)
+                # don't store request for other exceptions, it can be anything
+                # but in that case request has already been stored in HAR file
+            raise
         return response
 
     def async_open(self, url, **kwargs):
@@ -565,7 +452,7 @@ class Browser(object):
         # in case we did not catch something that should be
         response.raise_for_status()
 
-    def build_request(self, url, referrer=None, data_encoding=None, **kwargs):
+    def build_request(self, url, *, referrer=None, data_encoding=None, **kwargs):
         """
         Does the same job as open(), but returns a Request without
         submitting it.
@@ -592,11 +479,21 @@ class Browser(object):
                 req.method = 'GET'
 
         # convert unicode strings to proper encoding
-        if isinstance(req.data, unicode) and data_encoding:
+        if isinstance(req.data, str) and data_encoding:
             req.data = req.data.encode(data_encoding)
         if isinstance(req.data, dict) and data_encoding:
-            req.data = OrderedDict([(k, v.encode(data_encoding) if isinstance(v, unicode) else v)
-                                    for k, v in req.data.items()])
+            encoded_data = OrderedDict()
+            for k, v in req.data.items():
+                if isinstance(v, str):
+                    encoded_v = v.encode(data_encoding)
+                elif isinstance(v, list):
+                    encoded_v = [
+                        element.encode(data_encoding) if isinstance(element, str) else element for element in v
+                    ]
+                else:
+                    encoded_v = v
+                encoded_data[k] = encoded_v
+            req.data = encoded_data
 
         if referrer is None:
             referrer = self.get_referrer(self.url, url)
@@ -834,6 +731,7 @@ class PagesBrowser(DomainBrowser):
     _urls = None
 
     def __init__(self, *args, **kwargs):
+        self._urls = OrderedDict()
         self.highlight_el = kwargs.pop('highlight_el', False)
         super(PagesBrowser, self).__init__(*args, **kwargs)
 
@@ -847,12 +745,39 @@ class PagesBrowser(DomainBrowser):
         attrs = [(attr, getattr(self, attr)) for attr in dir(self) if not is_property(attr)]
         attrs = [v for v in attrs if isinstance(v[1], URL)]
         attrs.sort(key=lambda v: v[1]._creation_counter)
-        self._urls = OrderedDict(deepcopy(attrs))
-        for k, v in self._urls.items():
+        for k, v in deepcopy(attrs):
+            self._urls[k] = v
             setattr(self, k, v)
         for url in self._urls.values():
             url.browser = self
 
+    def __setattr__(self, key, value):
+        if isinstance(self._urls, OrderedDict):
+            # _urls is instanciated, we can now feed it accordingly.
+            if isinstance(value, URL):
+                # We want to either replace in-place, or add to the URLs.
+                if key in self._urls:
+                    # We want to actually make the old URL unusable.
+                    self._urls[key].browser = None
+
+                value = copy(value)
+                value.browser = self
+                self._urls[key] = value
+            elif key in self._urls:
+                # We want to remove the URL from our mapping only.
+                url = self._urls.pop(key)
+                url.browser = None
+
+        super().__setattr__(key, value)
+
+    def __delattr__(self, key):
+        if isinstance(self._urls, OrderedDict):
+            if key in self._urls:
+                # We want to remove the URL from our mapping.
+                del self._urls[key]
+
+        super().__delattr__(key)
+
     def open(self, *args, **kwargs):
         """
         Same method than
@@ -1013,12 +938,12 @@ class LoginBrowser(PagesBrowser):
         self.session.cookies.clear()
 
 
-class StatesMixin(object):
+class StatesMixin:
     """
     Mixin to store states of browser.
     """
 
-    __states__ = []
+    __states__ = ()
     """
     Saved state variables.
     """
@@ -1044,20 +969,22 @@ class StatesMixin(object):
         try:
             jcookies = json.loads(uncompressed)
         except ValueError:
-            try:
-                self.session.cookies = pickle.loads(uncompressed)
-            except (TypeError, EOFError, ValueError):
-                self.logger.error('Unable to reload cookies from storage')
-            else:
-                self.logger.warning('Reloaded deprecated cookie format')
+            self.logger.error('Unable to reload cookies from storage')
         else:
             for jcookie in jcookies:
                 self.session.cookies.set(**jcookie)
             self.logger.debug('Reloaded cookies from storage')
 
     def load_state(self, state):
-        if state.get('expire') and parser.parse(state['expire']) < datetime.now():
-            return self.logger.info('State expired, not reloading it from storage')
+        expire = state.get('expire')
+
+        if expire:
+            expire = parser.parse(expire)
+            if not expire.tzinfo:
+                expire = expire.replace(tzinfo=tz.tzlocal())
+            if expire < now_as_utc():
+                self.logger.info('State expired, not reloading it from storage')
+                return
 
         if 'cookies' in state:
             self._load_cookies(state['cookies'])
@@ -1070,7 +997,7 @@ class StatesMixin(object):
             self.locate_browser(state)
 
     def get_expire(self):
-        return unicode((datetime.now() + timedelta(minutes=self.STATE_DURATION)).replace(microsecond=0))
+        return str((now_as_utc() + timedelta(minutes=self.STATE_DURATION)).replace(microsecond=0))
 
     def dump_state(self):
         state = {}
@@ -1102,7 +1029,7 @@ class APIBrowser(DomainBrowser):
     """
 
     def build_request(self, *args, **kwargs):
-        if 'data' in kwargs:
+        if 'data' in kwargs and isinstance(kwargs['data'], dict):
             kwargs['data'] = json.dumps(kwargs['data'])
         if 'headers' not in kwargs:
             kwargs['headers'] = {}
@@ -1137,53 +1064,36 @@ class AbstractBrowserMissingParentError(Exception):
     pass
 
 
-class AbstractBrowser(Browser):
-    """ AbstractBrowser allow inheritance of a browser defined in another module.
+class MetaBrowser(type):
+    # we can remove this class as soon as we get rid of Abstract*
 
-    Websites can share many pages and code base. This class allow to load a browser
-    provided by another module and to build our own browser on top of it (like standard
-    python inheritance. Woob will install and download the PARENT module for you.
+    _parent_attr_re = re.compile(r'^[^.]+\.(.*)\.([^.]+)$')
 
-    PARENT is a mandatory attribute, it's the name of the module providing the parent Browser
+    def __new__(mcs, name, bases, dct):
+        from woob.tools.backend import Module  # Here to avoid file wide circular dependency
 
-    PARENT_ATTR is an optional attribute used when the parent module does not have only one
-    browser defined as BROWSER class attribute: you can customized the path of the object to load.
+        if name != 'AbstractBrowser' and AbstractBrowser in bases:
+            parent_attr = dct.get('PARENT_ATTR')
+            if parent_attr:
+                m = MetaBrowser._parent_attr_re.match(parent_attr)
+                path, klass_name = m.group(1, 2)
+                module = importlib.import_module('woob_modules.%s.%s' % (dct['PARENT'], path))
+                klass = getattr(module, klass_name)
+            else:
+                module = importlib.import_module('woob_modules.%s' % dct['PARENT'])
+                for attrname in dir(module):
+                    attr = getattr(module, attrname)
+                    if isinstance(attr, type) and issubclass(attr, Module) and attr != Module:
+                        klass = attr.BROWSER
+                        break
 
-    Note that you must pass a valid woob instance as first argument of the constructor.
-    """
-    PARENT = None
-    PARENT_ATTR = None
+            bases = tuple(klass if isinstance(base, mcs) else base for base in bases)
 
-    @classmethod
-    def _resolve_abstract(cls, woob):
-        if cls.PARENT is None:
-            raise AbstractBrowserMissingParentError("PARENT is not defined for browser %s" % cls)
+        return super(MetaBrowser, mcs).__new__(mcs, name, bases, dct)
 
-        try:
-            module = woob.load_or_install_module(cls.PARENT)
-        except ModuleInstallError as err:
-            raise ModuleInstallError('This module depends on %s module but %s\'s installation failed with: %s' % (cls.PARENT, cls.PARENT, err))
 
-        if cls.PARENT_ATTR is None:
-            parent = module.klass.BROWSER
-        else:
-            parent = reduce(getattr, cls.PARENT_ATTR.split('.'), module)
-
-        if parent is None:
-            raise AbstractBrowserMissingParentError("Failed to load parent class")
-
-        # Parent may be an AbstractBrowser as well
-        if hasattr(parent, '_resolve_abstract'):
-            parent._resolve_abstract(woob)
-
-        cls.__bases__ = (parent,)
-        cls.woob = woob
-
-    def __new__(cls, *args, **kwargs):
-        woob = kwargs.get("woob", kwargs.get("weboob"))
-        assert woob
-        cls._resolve_abstract(woob)
-        return Browser.__new__(cls, *args, **kwargs)
+class AbstractBrowser(metaclass=MetaBrowser):
+    """Don't use this class, import woob_modules.other_module.etc instead"""
 
 
 class OAuth2Mixin(StatesMixin):
@@ -1200,10 +1110,13 @@ class OAuth2Mixin(StatesMixin):
     token_type = None
     refresh_token = None
     oauth_state = None
+    authorized_date = None
 
     def __init__(self, *args, **kwargs):
         super(OAuth2Mixin, self).__init__(*args, **kwargs)
-        self.__states__ += ('access_token', 'access_token_expire', 'refresh_token', 'token_type')
+        self.__states__ += (
+            'access_token', 'refresh_token', 'token_type',
+        )
 
     def build_request(self, *args, **kwargs):
         headers = kwargs.setdefault('headers', {})
@@ -1212,12 +1125,28 @@ class OAuth2Mixin(StatesMixin):
         return super(OAuth2Mixin, self).build_request(*args, **kwargs)
 
     def dump_state(self):
-        self.access_token_expire = unicode(self.access_token_expire) if self.access_token_expire else None
-        return super(OAuth2Mixin, self).dump_state()
+        state = super(OAuth2Mixin, self).dump_state()
+
+        if self.authorized_date:
+            state['authorized_date'] = str(self.authorized_date)
+        if self.access_token_expire:
+            state['access_token_expire'] = self.access_token_expire.isoformat()
+
+        return state
 
     def load_state(self, state):
+        def load_date_or_none(date_str):
+            if date_str:
+                date_converted = parser.parse(date_str)
+                if isinstance(date_converted, datetime):
+                    return date_converted
+            return None
+
         super(OAuth2Mixin, self).load_state(state)
-        self.access_token_expire = parser.parse(self.access_token_expire) if self.access_token_expire else None
+        self.authorized_date = load_date_or_none(state.get('authorized_date'))
+        self.access_token_expire = load_date_or_none(state.get('access_token_expire'))
+        if self.access_token_expire and not self.access_token_expire.tzinfo:
+            self.access_token_expire = self.access_token_expire.replace(tzinfo=tz.tzlocal())
 
     def raise_for_status(self, response):
         if response.status_code == 401:
@@ -1227,7 +1156,8 @@ class OAuth2Mixin(StatesMixin):
 
     @property
     def logged(self):
-        return self.access_token is not None and (not self.access_token_expire or self.access_token_expire > datetime.now())
+        # 'access_token_expire' is already set as UTC in OAuth2Mixin.update_token.
+        return self.access_token is not None and (not self.access_token_expire or self.access_token_expire > now_as_utc())
 
     def do_login(self):
         if self.refresh_token:
@@ -1259,17 +1189,35 @@ class OAuth2Mixin(StatesMixin):
         raise BrowserRedirect(self.build_authorization_uri())
 
     def handle_callback_error(self, values):
-        # Here we try to catch callback errors occurring during enrollment
-        # Ideally overload this method in each module to catch specific error
-        assert values.get('code'), "No 'code' was found into the callback url, please raise the right error: %s" % values
+        wrongpass_on_webauth_errors = re.compile('|'.join((
+            'operation canceled by the client',
+            'login cancelled',
+            'consent denied',
+            'psu cancelled the transaction',
+        )))
+
+        error = values.get('error')
+        error_message = values.get('error_description')
+
+        if error == 'access_denied':
+            if error_message:
+                if wrongpass_on_webauth_errors.search(error_message.lower()):
+                    raise BrowserIncorrectPassword(error_message)
+                raise AssertionError(f'Unhandled callback error_message: {error_message}')
+            # access_denied with no error message
+            raise BrowserIncorrectPassword()
+        if error == 'server_error':
+            raise BrowserUnavailable()
+        raise AssertionError(f'Unhandled callback error: {error}, error_message: {error_message}')
 
     def build_access_token_parameters(self, values):
-        return {'code':             values['code'],
-                'grant_type':       'authorization_code',
-                'redirect_uri':     self.redirect_uri,
-                'client_id':        self.client_id,
-                'client_secret':    self.client_secret,
-                }
+        return {
+            'code': values['code'],
+            'grant_type': 'authorization_code',
+            'redirect_uri': self.redirect_uri,
+            'client_id': self.client_id,
+            'client_secret': self.client_secret,
+        }
 
     def do_token_request(self, data):
         return self.open(self.ACCESS_TOKEN_URI, data=data)
@@ -1281,13 +1229,14 @@ class OAuth2Mixin(StatesMixin):
             values = auth_uri
         else:
             values = dict(parse_qsl(urlparse(auth_uri).query))
-        self.handle_callback_error(values)
+        if not values.get('code'):
+            self.handle_callback_error(values)
+        self.authorized_date = now_as_utc()
         data = self.build_access_token_parameters(values)
         try:
             auth_response = self.do_token_request(data).json()
         except ClientError:
-            self.logger.debug('Failed to request access token')
-            raise BrowserIncorrectPassword()
+            raise AssertionError('PSU has successfully logged in, but the request to get an access token failed.')
 
         self.update_token(auth_response)
 
@@ -1317,7 +1266,7 @@ class OAuth2Mixin(StatesMixin):
         if 'refresh_token' in auth_response:
             self.refresh_token = auth_response['refresh_token']
         self.access_token = auth_response['access_token']
-        self.access_token_expire = datetime.now() + timedelta(seconds=int(auth_response['expires_in']))
+        self.access_token_expire = now_as_utc() + timedelta(seconds=int(auth_response['expires_in']))
 
 
 class OAuth2PKCEMixin(OAuth2Mixin):
@@ -1337,140 +1286,21 @@ class OAuth2PKCEMixin(OAuth2Mixin):
 
     def build_authorization_parameters(self):
         params = {
-            'redirect_uri':    self.redirect_uri,
+            'redirect_uri': self.redirect_uri,
             'code_challenge_method': 'S256',
-            'code_challenge':  self.pkce_challenge,
-            'client_id':       self.client_id,
+            'code_challenge': self.pkce_challenge,
+            'client_id': self.client_id,
         }
         if self.oauth_state:
             params['state'] = self.oauth_state
         return params
 
     def build_access_token_parameters(self, values):
-        return {'code':             values['code'],
-                'grant_type':       'authorization_code',
-                'code_verifier':    self.pkce_verifier,
-                'redirect_uri':     self.redirect_uri,
-                'client_id':        self.client_id,
-                'client_secret':    self.client_secret,
-                }
-
-
-class TwoFactorBrowser(LoginBrowser, StatesMixin):
-    # period to keep the same state
-    # it is different from STATE_DURATION which updates the expire date at each dump
-    TWOFA_DURATION = None
-
-    INTERACTIVE_NAME = 'request_information'
-    # dict of config keys and methods used for double authentication
-    # must be set up in the init to handle function pointers
-    AUTHENTICATION_METHODS = {}
-
-    # list of cookie keys to clear before dumping state
-    COOKIES_TO_CLEAR = ()
-
-    # login can also be done with credentials without 2FA
-    HAS_CREDENTIALS_ONLY = False
-
-    def __init__(self, config, *args, **kwargs):
-        super(TwoFactorBrowser, self).__init__(*args, **kwargs)
-        self.config = config
-        self.is_interactive = config.get(self.INTERACTIVE_NAME, Value()).get() is not None
-        self.twofa_logged_date = None
-        self.__states__ += ('twofa_logged_date',)
-
-    def get_expire(self):
-        if self.twofa_logged_date:
-            logged_date = parser.parse(self.twofa_logged_date)
-        else:
-            logged_date = None
-
-        expires_dates = [datetime.now() + timedelta(minutes=self.STATE_DURATION)]
-        if logged_date and self.TWOFA_DURATION is not None:
-            expires_dates.append(logged_date + timedelta(minutes=self.TWOFA_DURATION))
-
-        return unicode(max(expires_dates).replace(microsecond=0))
-
-    def dump_state(self):
-        self.clear_not_2fa_cookies()
-        # so the date can be parsed in json
-        # because twofa_logged_date is in state
-        if self.twofa_logged_date:
-            self.twofa_logged_date = str(self.twofa_logged_date)
-        return super(TwoFactorBrowser, self).dump_state()
-
-    def init_login(self):
-        """
-        Abstract method to implement initiation of login on website.
-
-        This method should raise an exception.
-
-        SCA exceptions :
-        - AppValidation for polling method
-        - BrowserQuestion for SMS method, token method etc.
-
-        Any other exceptions, default to BrowserIncorrectPassword.
-        """
-        raise NotImplementedError()
-
-    def clear_init_cookies(self):
-        # clear cookies to avoid some errors
-        self.session.cookies.clear()
-
-    def clear_not_2fa_cookies(self):
-        # clear cookies that we don't need for 2FA
-        for cookie_key in self.COOKIES_TO_CLEAR:
-            if cookie_key in self.session.cookies:
-                del self.session.cookies[cookie_key]
-
-    def check_interactive(self):
-        if not self.is_interactive:
-            raise NeedInteractiveFor2FA()
-
-    def do_double_authentication(self):
-        """
-        This method will check AUTHENTICATION_METHODS
-        to dispatch to the right handle_* method.
-
-        If no backend configuration could be found,
-        it will then call init_login method.
-        """
-
-        def clear_sca_key(config_key):
-            if self.config.get(config_key):
-                self.config[config_key] = self.config[config_key].default
-
-        assert self.AUTHENTICATION_METHODS, 'There is no config for the double authentication.'
-        self.twofa_logged_date = None
-
-        for config_key, handle_method in self.AUTHENTICATION_METHODS.items():
-            config_value = self.config.get(config_key, Value())
-            if not config_value:
-                continue
-
-            setattr(self, config_key, config_value.get())
-            if getattr(self, config_key):
-                try:
-                    handle_method()
-                except BrowserInteraction:
-                    # If a BrowserInteraction is raised during the handling of the sca_key,
-                    # we need to clear it before restarting the process to prevent it to block
-                    # other sca_keys handling.
-                    clear_sca_key(config_key)
-                    raise
-
-                self.twofa_logged_date = datetime.now()
-
-                # cleaning authentication config keys
-                for config_key in self.AUTHENTICATION_METHODS.keys():
-                    clear_sca_key(config_key)
-
-                break
-        else:
-            if not self.HAS_CREDENTIALS_ONLY:
-                self.check_interactive()
-
-            self.clear_init_cookies()
-            self.init_login()
-
-    do_login = do_double_authentication
+        return {
+            'code': values['code'],
+            'grant_type': 'authorization_code',
+            'code_verifier': self.pkce_verifier,
+            'redirect_uri': self.redirect_uri,
+            'client_id': self.client_id,
+            'client_secret': self.client_secret,
+        }
