@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2014 Romain Bignon
 #
 # This file is part of woob.
@@ -17,8 +15,7 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
+import importlib
 import os
 import re
 import sys
@@ -29,12 +26,12 @@ import traceback
 import lxml.html
 
 from woob.tools.log import getLogger, DEBUG_FILTERS
-from woob.tools.compat import basestring, unicode, with_metaclass
 from woob.browser.pages import NextPage
 from woob.capabilities.base import FetchError
 
 from .filters.standard import _Filter, CleanText
 from .filters.html import AttributeNotFound, XPathNotFound
+from .filters.json import Dict
 
 
 __all__ = [
@@ -96,9 +93,35 @@ def method(klass):
     return inner
 
 
-class AbstractElement(object):
+class AbstractElement:
     _creation_counter = 0
+
     condition = None
+    """The condition to parse the element.
+
+    This allows ignoring certain elements if certain fields are not valid,
+    or if the element should actually be parsed using another class.
+
+    This property can be defined as:
+
+    * None or True, to signify that the element should be parsed regardless.
+    * False, to signify that the element should not be parsed regardless.
+    * A filter returning a falsy or non-falsy object, evaluated with the
+      constructed document section (HTML element or JSON data) for the element.
+    * A method returning a falsy or non-falsy object, evaluated with the
+      element object directly.
+    """
+
+    def __new__(cls, *args, **kwargs):
+        """ Accept any arguments, necessary for ItemElementFromAbstractPage __new__
+        override.
+
+        ItemElementFromAbstractPage, in its overridden __new__, removes itself from
+        class hierarchy so its __new__ is called only once. In python 3, default
+        (object) __new__ is then used for next instantiations but it's a slot/fixed
+        version supporting only one argument (type to instanciate).
+        """
+        return object.__new__(cls)
 
     def __init__(self, page, parent=None, el=None):
         self.page = page
@@ -165,6 +188,25 @@ class AbstractElement(object):
         else:
             self.env = deepcopy(page.params)
 
+    def check_condition(self):
+        """Get whether our condition is respected or not."""
+        if self.condition is None or self.condition is True:
+            return True
+        elif self.condition is False:
+            return False
+        elif isinstance(self.condition, _Filter):
+            if self.condition(self.el):
+                return True
+        elif callable(self.condition):
+            if self.condition():
+                return True
+        else:
+            assert isinstance(self.condition, str)
+            if self.el.xpath(self.condition):
+                return True
+
+        return False
+
 
 class ListElement(AbstractElement):
     item_xpath = None
@@ -200,7 +242,7 @@ class ListElement(AbstractElement):
             yield self.el
 
     def __iter__(self):
-        if self.condition is not None and not self.condition():
+        if not self.check_condition():
             return
 
         self.parse(self.el)
@@ -211,7 +253,7 @@ class ListElement(AbstractElement):
                 attr = getattr(self, attrname)
                 if isinstance(attr, type) and issubclass(attr, AbstractElement) and attr != type(self):
                     item = attr(self.page, self, el)
-                    if item.condition is not None and not item.condition():
+                    if not item.check_condition():
                         continue
 
                     item.handle_loaders()
@@ -287,14 +329,31 @@ class _ItemElementMeta(type):
         return new_class
 
 
-class ItemElement(with_metaclass(_ItemElementMeta, AbstractElement)):
+class ItemElementRerootMixin:
+    """
+    Mixin used to reroot an ItemElement by defining a reroot_xpath.
+    """
+
+    reroot_xpath = None
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        if self.reroot_xpath:
+            if hasattr(self.el, 'xpath'):
+                self.el = self.el.xpath(self.reroot_xpath)
+            elif isinstance(self.el, (dict, list)):
+                self.el = Dict.select(self.reroot_xpath.split('/'), self)
+
+
+class ItemElement(AbstractElement, metaclass=_ItemElementMeta):
     _attrs = None
     _loaders = None
     klass = None
     validate = None
     skip_optional_fields_errors = False
 
-    class Index(object):
+    class Index:
         pass
 
     def __init__(self, *args, **kwargs):
@@ -337,15 +396,18 @@ class ItemElement(with_metaclass(_ItemElementMeta, AbstractElement)):
             fd.write(html)
         self.logger.debug('highlighted object to %s', fn)
 
-    def __call__(self, obj=None):
+    def __call__(self, obj=None, **kwargs):
         if obj is not None:
             self.obj = obj
 
+        for key, value in kwargs.items():
+            self.env[key] = value
+
         for obj in self:
             return obj
 
     def __iter__(self):
-        if self.condition is not None and not self.condition():
+        if not self.check_condition():
             return
 
         highlight = False
@@ -389,7 +451,7 @@ class ItemElement(with_metaclass(_ItemElementMeta, AbstractElement)):
                 raise
             else:
                 value = FetchError
-        logger = getLogger('b2filters')
+        logger = getLogger('woob.browser.b2filters')
         logger.log(DEBUG_FILTERS, "%s.%s = %r" % (self._random_id, key, value))
         setattr(self.obj, key, value)
 
@@ -398,56 +460,41 @@ class ItemElementFromAbstractPageError(Exception):
     pass
 
 
-class ItemElementFromAbstractPage(ItemElement):
-    """
-    ItemElementFromAbstractPage allow to use ItemElement class
-    from an AbstractPage of another module and to overwrite specific obj_*.
+class MetaAbstractItemElement(type):
+    # hope we get rid of this real fast
+    def __new__(mcs, name, bases, dct):
+        from woob.tools.backend import Module  # here to avoid file wide circular dependency
 
-    ITER_ELEMENT is an optional attribute to define which class should be used
-    from the page loaded by AbstractPage, default to current iter_element class name.
+        if name != 'ItemElementFromAbstractPage' and ItemElementFromAbstractPage in bases:
+            parent_attr = dct.get('BROWSER_ATTR', None)
+            if parent_attr:
+                m = re.match(r'^[^.]+\.(.*)\.([^.]+)$', parent_attr)
+                path, klass_name = m.group(1, 2)
+                module = importlib.import_module('woob_modules.%s.%s' % (dct['PARENT'], path))
+                browser_klass = getattr(module, klass_name)
+            else:
+                module = importlib.import_module('woob_modules.%s' % dct['PARENT'])
+                for attrname in dir(module):
+                    attr = getattr(module, attrname)
+                    if isinstance(attr, type) and issubclass(attr, Module) and attr != Module:
+                        browser_klass = attr.BROWSER
+                        break
 
-    For example:
+            url = getattr(browser_klass, dct['PARENT_URL'])
+            page_class = url.klass
 
-    >>> from woob.browser.elements import DictElement, ItemElementFromAbstractPage, method
-    >>> from woob.browser.pages import AbstractPage
-    >>>
-    >>> class AccountsPage(AbstractPage):
-    ...     PARENT = 'stet'
-    ...     PARENT_URL = 'accounts'
-    ...
-    ...     @method
-    ...     class iter_accounts(DictElement):
-    ...         item_xpath = 'accounts'
-    ...
-    ...         class item(ItemElementFromAbstractPage):
-    ...             ITER_ELEMENT = "iter_accounts"  # optional
-    ...
-    ...             obj_label = "XXX"
-    """
+            element_class = getattr(page_class, dct['ITER_ELEMENT']).klass
+            item_class = element_class.item
 
-    PARENT = None
-    PARENT_URL = None
-    ITER_ELEMENT = None
+            bases = tuple(item_class if isinstance(base, mcs) else base for base in bases)
 
-    @classmethod
-    def _resolve_abstract(cls, page, iter_element, *args, **kwargs):
-        if not (cls.PARENT and cls.PARENT_URL and cls.ITER_ELEMENT):
-            raise ItemElementFromAbstractPageError('page %s is not an AbstractPage' % page)
+            return type(name, bases, dct)
 
-        abstract_page = page.__class__.__bases__[0]
-        parent_iter_element = getattr(abstract_page, cls.ITER_ELEMENT, None)
-        if not parent_iter_element:
-            raise ItemElementFromAbstractPageError(
-                "%s has no iter_element '%s'" % (abstract_page, cls.ITER_ELEMENT)
-            )
+        return super(MetaAbstractItemElement, mcs).__new__(mcs, name, bases, dct)
 
-        cls._attrs = list(set(iter_element.item._attrs + parent_iter_element.klass.item._attrs))
-        cls.__bases__ += (parent_iter_element.klass.item,)
 
-    def __new__(cls, *args, **kwargs):
-        if len(cls.__bases__) == 1:
-            cls._resolve_abstract(*args, **kwargs)
-        return object.__new__(cls)
+class ItemElementFromAbstractPage(metaclass=MetaAbstractItemElement):
+    """Don't use this class, import woob_modules.other_module.etc instead"""
 
 
 class TableElement(ListElement):
@@ -466,7 +513,7 @@ class TableElement(ListElement):
                 cols = getattr(self, attrname)
                 if not isinstance(cols, (list,tuple)):
                     cols = [cols]
-                columns[m.group(1)] = [s.lower() if isinstance(s, (str, unicode)) else s for s in cols]
+                columns[m.group(1)] = [s.lower() if isinstance(s, str) else s for s in cols]
 
         colnum = 0
         for el in self.el.xpath(self.head_xpath):
@@ -474,7 +521,7 @@ class TableElement(ListElement):
             for name, titles in columns.items():
                 if name in self._cols:
                     continue
-                if title.lower() in [s for s in titles if isinstance(s, (str, unicode))] or \
+                if title.lower() in [s for s in titles if isinstance(s, str)] or \
                    any(map(lambda x: x.match(title), [s for s in titles if isinstance(s, type(re.compile('')))])):
                     self._cols[name] = colnum
             try:
@@ -491,7 +538,7 @@ class DictElement(ListElement):
         if self.item_xpath is None:
             selector = []
 
-        elif isinstance(self.item_xpath, basestring):
+        elif isinstance(self.item_xpath, str):
             selector = self.item_xpath.split('/')
 
         else:
@@ -505,8 +552,10 @@ class DictElement(ListElement):
                 bases = [el[int(key)] if isinstance(el, list) else el[key] for el in bases]
 
         for base in bases:
-            for el in base:
-                yield el
+            if isinstance(base, dict):
+                yield from base.values()
+            else:
+                yield from base
 
 
 def magic_highlight(els, open_browser=True):
@@ -531,7 +580,7 @@ def magic_highlight(els, open_browser=True):
     for el in els:
         el.attrib['style'] = saved[el]
 
-    _, fn = tempfile.mkstemp(dir='/tmp', prefix='woob-highlight', suffix='.html')
+    _, fn = tempfile.mkstemp(prefix='woob-highlight', suffix='.html')
     with open(fn, 'w') as fd:
         fd.write(html)
 
