@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2016      Vincent A
 #
 # This file is part of woob.
@@ -26,8 +24,6 @@
 # import certificate:
 #   find -L /etc/ssl/certs -name "*.pem" | while read f; do certutil -A -d pki -i $f -n $f -t TCu,Cu,Tu; done
 
-from __future__ import absolute_import
-
 from functools import wraps
 from io import RawIOBase, BufferedRWPair
 import hashlib
@@ -56,7 +52,7 @@ __all__ = ['init_nss', 'inject_in_urllib3', 'certificate_db_filename']
 CTX = None
 INIT_PID = None
 INIT_ARGS = None
-LOGGER = getLogger('woob.browser.nss')
+LOGGER = getLogger(__name__)
 
 
 def nss_version():
@@ -183,7 +179,7 @@ class NSSFile(RawIOBase):
     writable = readable
 
 
-class Wrapper(object):
+class Wrapper:
     def __init__(self, obj):
         self.__obj = obj
         self.__timeout = nss.io.PR_INTERVAL_NO_TIMEOUT
