@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2014 Romain Bignon
 #
 # This file is part of woob.
@@ -17,22 +15,18 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import absolute_import
-
+import codecs
+import importlib
+import re
+import warnings
 from collections import OrderedDict
 from functools import wraps
-import warnings
 from io import BytesIO, StringIO
-import codecs
-from cgi import parse_header
-from functools import reduce
-import re
-import sys
+from urllib.parse import urljoin
 
 import requests
 
-from woob.exceptions import ParseError, ModuleInstallError
-from woob.tools.compat import basestring, unicode, urljoin
+from woob.exceptions import ParseError
 from woob.tools.json import json, mini_jsonpath
 from woob.tools.log import getLogger
 from woob.tools.pdf import decompress_pdf
@@ -101,7 +95,7 @@ class NextPage(Exception):
         self.request = request
 
 
-class Page(object):
+class Page:
     """
     Represents a page.
 
@@ -119,7 +113,7 @@ class Page(object):
     :param params: optional dictionary containing parameters given to the page (see :class:`woob.browser.url.URL`)
     :type params: :class:`dict`
     :param encoding: optional parameter to force the encoding of the page, overrides :attr:`ENCODING`
-    :type encoding: :class:`basestring`
+    :type encoding: :class:`str`
 
     """
 
@@ -129,6 +123,25 @@ class Page(object):
     It is recommended to use None for autodetection.
     """
 
+    is_here = None
+    """The condition to verify that the page corresponds to the response.
+
+    This allows having different pages on equivalent or conflicting URL
+    patterns identified using the response's method, URL, headers, or content,
+    by defining is_here on pages associated with such patterns.
+
+    This property can be defined as:
+
+    * None or True, to signify that the page should be matched regardless
+      of the response.
+    * False, to signify that the page should not be matched regardless of
+      the response.
+    * A filter returning a falsy or non-falsy object, evaluated with the
+      constructed document for the page.
+    * A method returning a falsy or non-falsy object, evaluated with the
+      page object directly.
+    """
+
     logged = False
     """
     If True, the page is in a restricted area of the website. Useful with
@@ -188,7 +201,7 @@ class Page(object):
     @property
     def text(self):
         """
-        Content of the response, in unicode, decoded with :attr:`encoding`.
+        Content of the response, in str, decoded with :attr:`encoding`.
         """
         return self.response.text
 
@@ -319,11 +332,11 @@ class Form(OrderedDict):
                 if len(options) == 0:
                     options = inp.xpath('.//option')
                 if len(options) == 0:
-                    value = u''
+                    value = ''
                 else:
-                    value = options[0].attrib.get('value', options[0].text or u'')
+                    value = options[0].attrib.get('value', options[0].text or '')
             else:
-                value = inp.attrib.get('value', inp.text or u'')
+                value = inp.attrib.get('value', inp.text or '')
             # TODO check if value already exists, emit warning
             self[name] = value
 
@@ -353,7 +366,7 @@ class Form(OrderedDict):
         Submit the form and tell browser to be located to the new page.
 
         :param data_encoding: force encoding used to submit form data (defaults to the current page encoding)
-        :type data_encoding: :class:`basestring`
+        :type data_encoding: :class:`str`
         """
         kwargs.setdefault('data_encoding', self.page.encoding)
         self.headers = kwargs.pop('headers', None)
@@ -401,9 +414,7 @@ class CsvPage(Page):
             encoding = 'utf-8'
         if self.NEWLINES_HACK:
             content = content.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
-        if sys.version_info.major > 2:
-            return self.parse(StringIO(content.decode(encoding)))
-        return self.parse(BytesIO(content), encoding)
+        return self.parse(StringIO(content.decode(encoding)))
 
     def parse(self, data, encoding=None):
         """
@@ -422,10 +433,7 @@ class CsvPage(Page):
         for i, row in enumerate(reader):
             if self.HEADER and i+1 < self.HEADER:
                 continue
-            if sys.version_info.major > 2:
-                row = [c.strip() for c in row]
-            else:
-                row = [c.strip() for c in self.decode_row(row, encoding)]
+            row = [c.strip() for c in row]
             if header is None and self.HEADER:
                 header = row
             else:
@@ -442,7 +450,7 @@ class CsvPage(Page):
         Method called by :meth:`CsvPage.parse` to decode a row using the given encoding.
         """
         if encoding:
-            return [unicode(cell, encoding) for cell in row]
+            return [str(cell, encoding) for cell in row]
         else:
             return row
 
@@ -528,13 +536,13 @@ class XMLPage(Page):
 
     def detect_encoding(self):
         import re
-        m = re.search(b'<\?xml version="1.0" encoding="(.*)"\?>', self.data)
+        m = re.search(br'<\?xml version="1.0" encoding="(.*)"\?>', self.data)
         if m:
             return self.normalize_encoding(m.group(1))
 
     def build_doc(self, content):
         import lxml.etree as etree
-        parser = etree.XMLParser(encoding=self.encoding)
+        parser = etree.XMLParser(encoding=self.encoding, resolve_entities=False)
         return etree.parse(BytesIO(content), parser)
 
 
@@ -558,7 +566,7 @@ class HTMLPage(Page):
     :param params: optional dictionary containing parameters given to the page (see :class:`woob.browser.url.URL`)
     :type params: :class:`dict`
     :param encoding: optional parameter to force the encoding of the page
-    :type encoding: :class:`basestring`
+    :type encoding: :class:`str`
 
     """
 
@@ -575,7 +583,7 @@ class HTMLPage(Page):
     Default value is None, means refreshes aren't handled.
     """
 
-    REFRESH_XPATH = '//head/meta[lower-case(@http-equiv)="refresh"]'
+    REFRESH_XPATH = '//head//meta[lower-case(@http-equiv)="refresh"]'
     """
     Default xpath, which is also the most commun, override it if needed
     """
@@ -586,10 +594,7 @@ class HTMLPage(Page):
     """
 
     def __init__(self, *args, **kwargs):
-        import lxml.html as html
-        ns = html.etree.FunctionNamespace(None)
-        self.define_xpath_functions(ns)
-
+        self.setup_xpath_functions()
         super(HTMLPage, self).__init__(*args, **kwargs)
 
     def on_load(self):
@@ -614,7 +619,15 @@ class HTMLPage(Page):
             else:
                 self.logger.debug('Do not refresh to %s because %s > REFRESH_MAX(%s)' % (url, sleep, self.REFRESH_MAX))
 
-    def define_xpath_functions(self, ns):
+    @classmethod
+    def setup_xpath_functions(cls):
+        import lxml.html as html
+
+        ns = html.etree.FunctionNamespace(None)
+        cls.define_xpath_functions(ns)
+
+    @classmethod
+    def define_xpath_functions(cls, ns):
         """
         Define XPath functions on the given lxml function namespace.
 
@@ -675,21 +688,25 @@ class HTMLPage(Page):
                     return nodes
             return []
 
+        def distinct_values(context, text):
+            return list(set(text))
+
         ns['has-class'] = has_class
         ns['starts-with'] = starts_with
         ns['ends-with'] = ends_with
         ns['matches'] = matches
         ns['first-non-empty'] = first_non_empty
+        ns['distinct-values'] = distinct_values
 
     def build_doc(self, content):
         """
         Method to build the lxml document from response and given encoding.
         """
         encoding = self.encoding
-        if encoding == u'latin-1':
-            encoding = u'latin1'
+        if encoding == 'latin-1':
+            encoding = 'latin1'
         if encoding:
-            encoding = encoding.replace(u'iso8859_', u'iso8859-')
+            encoding = encoding.replace('iso8859_', 'iso8859-')
         import lxml.html as html
         parser = html.HTMLParser(encoding=encoding)
         doc = html.parse(BytesIO(content), parser, base_url=self.url)
@@ -706,20 +723,25 @@ class HTMLPage(Page):
         encoding = self.encoding
         for content in self.doc.xpath('//head/meta[lower-case(@http-equiv)="content-type"]/@content'):
             # meta http-equiv=content-type content=...
-            _, params = parse_header(content)
-            if 'charset' in params:
-                encoding = self.normalize_encoding(params['charset'].strip("'\""))
+
+            # Use request's method to get encoding from headers, so we simulate
+            # an headers dict.
+            encoding = self.normalize_encoding(
+                requests.utils.get_encoding_from_headers(
+                    {'content-type': content}
+                )
+            )
 
         for charset in self.doc.xpath('//head/meta[@charset]/@charset'):
             # meta charset=...
             encoding = self.normalize_encoding(charset)
 
-        if encoding == u'iso-8859-1' or not encoding:
-            encoding = u'windows-1252'
+        if encoding == 'iso-8859-1' or not encoding:
+            encoding = 'windows-1252'
         try:
             codecs.lookup(encoding)
         except LookupError:
-            encoding = u'windows-1252'
+            encoding = 'windows-1252'
 
         return encoding
 
@@ -753,7 +775,7 @@ class HTMLPage(Page):
                 i += 1
                 continue
 
-            if isinstance(submit, basestring):
+            if isinstance(submit, str):
                 submit_el = el.xpath(submit)[0]
             else:
                 submit_el = submit
@@ -809,7 +831,7 @@ class GWTPage(Page):
                 array = el
             elif array and isinstance(el, int) and len(array) >= el >= 1:
                 doc.append(array[el - 1])
-            elif array and isinstance(el, basestring):
+            elif array and isinstance(el, str):
                 doc.append(self.get_date(el))
         return doc
 
@@ -843,12 +865,12 @@ class PDFPage(Page):
         try:
             doc = decompress_pdf(content)
         except OSError as e:
-            raise ParseError(u'Make sure mupdf-tools is installed (%s)' % e)
+            raise ParseError('Make sure mupdf-tools is installed (%s)' % e)
 
         return doc
 
 
-class LoggedPage(object):
+class LoggedPage:
     """
     A page that only logged users can reach. If we did not get a redirection
     for this page, we are sure that the login is still active.
@@ -859,7 +881,7 @@ class LoggedPage(object):
     logged = True
 
 
-class ChecksumPage(object):
+class ChecksumPage:
     """
     Compute a checksum of raw content before parsing it.
     """
@@ -877,55 +899,39 @@ class AbstractPageError(Exception):
     pass
 
 
-class AbstractPage(Page):
-    PARENT = None
-    PARENT_URL = None
-    BROWSER_ATTR = None
+class MetaPage(type):
+    # we can remove this class as soon as we get rid of Abstract*
+    def __new__(mcs, name, bases, dct):
+        from woob.tools.backend import Module  # here to avoid file wide circular dependency
 
-    @classmethod
-    def _resolve_abstract(cls, browser):
-        try:
-            woob = browser.woob
-        except AttributeError:
-            try:
-                woob = browser.weboob
-            except AttributeError:
-                woob = None
-        if not woob:
-            raise AbstractPageError("woob is not defined in %s" % browser)
+        if name != 'AbstractPage' and AbstractPage in bases:
+            parent_attr = dct.get('BROWSER_ATTR')
+            if parent_attr:
+                m = re.match(r'^[^.]+\.(.*)\.([^.]+)$', parent_attr)
+                path, klass_name = m.group(1, 2)
+                module = importlib.import_module('woob_modules.%s.%s' % (dct['PARENT'], path))
+                browser_klass = getattr(module, klass_name)
+            else:
+                module = importlib.import_module('woob_modules.%s' % dct['PARENT'])
+                for attrname in dir(module):
+                    attr = getattr(module, attrname)
+                    if isinstance(attr, type) and issubclass(attr, Module) and attr != Module:
+                        browser_klass = attr.BROWSER
+                        break
 
-        if cls.PARENT is None:
-            raise AbstractPageError("PARENT is not defined for page %s" % cls.__name__)
+            url = getattr(browser_klass, dct['PARENT_URL'])
+            klass = url.klass
 
-        if cls.PARENT_URL is None:
-            raise AbstractPageError("PARENT_URL is not defined for page %s" % cls.__name__)
+            bases = tuple(klass if isinstance(base, mcs) else base for base in bases)
 
-        try:
-            parent_module = woob.load_or_install_module(cls.PARENT)
-        except ModuleInstallError as err:
-            raise ModuleInstallError('This module depends on %s module but %s\'s installation failed with: %s' % (cls.PARENT, cls.PARENT, err))
-
-        if cls.BROWSER_ATTR is None:
-            parent_browser = parent_module.klass.BROWSER
-        else:
-            parent_browser = reduce(getattr, cls.BROWSER_ATTR.split('.'), parent_module)
-
-        parent = getattr(parent_browser, cls.PARENT_URL, None)
-        if parent is None:
-            raise AbstractPageError("cls.PARENT_URL is not defined in %s" % browser)
-
-        # Parent may be an AbstractPage as well
-        if hasattr(parent.klass, '_resolve_abstract'):
-            parent.klass._resolve_abstract(browser)
-
-        cls.__bases__ = (parent.klass,)
-
-    def __new__(cls, browser, *args, **kwargs):
-        cls._resolve_abstract(browser)
-        return object.__new__(cls)
+        return super(MetaPage, mcs).__new__(mcs, name, bases, dct)
 
 
-class LoginPage(object):
+class AbstractPage(metaclass=MetaPage):
+    """Don't use this class, import woob_modules.other_module.etc instead"""
+
+
+class LoginPage:
     def on_load(self):
         if not self.browser.logging_in:
             raise LoggedOut()
