@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2018      Vincent Ardisson
 #
 # This file is part of woob.
@@ -17,8 +15,6 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import unicode_literals, absolute_import
-
 import codecs
 from collections import OrderedDict
 from contextlib import contextmanager
@@ -29,6 +25,10 @@ import hashlib
 from tempfile import NamedTemporaryFile
 import time
 import logging
+from urllib.parse import (
+    urljoin, urlparse, urlencode, parse_qsl,
+    urlunparse,
+)
 
 try:
     from selenium import webdriver
@@ -46,10 +46,6 @@ from selenium.webdriver.common.by import By
 from selenium.webdriver.firefox.firefox_profile import FirefoxProfile
 from selenium.webdriver.remote.command import Command
 from woob.tools.log import getLogger
-from woob.tools.compat import (
-    urljoin, urlparse, urlencode, parse_qsl,
-    urlunparse,
-)
 
 from .pages import HTMLPage as BaseHTMLPage
 from .url import URL
@@ -64,7 +60,7 @@ __all__ = (
 )
 
 
-class CustomCondition(object):
+class CustomCondition:
     """Abstract condition class
 
     In Selenium, waiting is done on callable objects named "conditions".
@@ -193,7 +189,7 @@ class StablePageCondition(CustomCondition):
     def __call__(self, driver):
         self._purge()
 
-        hashed = hashlib.md5(driver.page_source.encode('utf-8')).hexdigest()
+        hashed = hashlib.new("md5", driver.page_source.encode('utf-8')).hexdigest()  # nosec
         now = time.time()
         page_id = driver.find_element_by_xpath('/*').id
 
@@ -250,7 +246,7 @@ def link_locator(text, partial=False):
         return (By.LINK_TEXT, text)
 
 
-class ElementWrapper(object):
+class ElementWrapper:
     """Wrapper to Selenium element to ressemble lxml.
 
     Some differences:
@@ -286,7 +282,7 @@ class ElementWrapper(object):
         return getattr(self.wrapped, attr)
 
     @property
-    class attrib(object):
+    class attrib:
         def __init__(self, el):
             self.el = el
 
@@ -303,7 +299,7 @@ class ElementWrapper(object):
             return v
 
 
-class SeleniumPage(object):
+class SeleniumPage:
     """Page to use in a SeleniumBrowser
 
     Differences with regular woob Pages:
@@ -355,14 +351,12 @@ class HTMLPage(BaseHTMLPage):
 OPTIONS_CLASSES = {
     webdriver.Firefox: webdriver.FirefoxOptions,
     webdriver.Chrome: webdriver.ChromeOptions,
-    webdriver.PhantomJS: webdriver.ChromeOptions, # unused, put dummy thing
 }
 
 
 CAPA_CLASSES = {
     webdriver.Firefox: DesiredCapabilities.FIREFOX,
     webdriver.Chrome: DesiredCapabilities.CHROME,
-    webdriver.PhantomJS: DesiredCapabilities.PHANTOMJS,
 }
 
 
@@ -377,7 +371,7 @@ class DirFirefoxProfile(FirefoxProfile):
         return super(DirFirefoxProfile, self)._create_tempfolder()
 
 
-class FakeResponse(object):
+class FakeResponse:
     page = None
 
     def __init__(self, **kwargs):
@@ -392,7 +386,7 @@ class SeleniumBrowserSetupError(Exception):
     """
 
 
-class SeleniumBrowser(object):
+class SeleniumBrowser:
     """Browser similar to PagesBrowser, but using Selenium.
 
     URLs instances can be used. The need_login decorator can be used too.
@@ -530,9 +524,10 @@ class SeleniumBrowser(object):
                 capa['profile'] = self.responses_dirname
             self.driver = self.DRIVER(options=options, capabilities=capa, **driver_kwargs)
         elif self.DRIVER is webdriver.Chrome:
+            if self.HEADLESS:
+                # Prevent random renderer timeout
+                options.add_argument('--disable-gpu')
             self.driver = self.DRIVER(options=options, desired_capabilities=capa, **driver_kwargs)
-        elif self.DRIVER is webdriver.PhantomJS:
-            self.driver = self.DRIVER(desired_capabilities=capa, **driver_kwargs)
         else:
             raise NotImplementedError()
 
@@ -649,7 +644,7 @@ class SeleniumBrowser(object):
         return ret
 
     def save_response_if_changed(self):
-        hash = hashlib.md5(self.driver.page_source.encode('utf-8')).hexdigest()
+        hash = hashlib.new("md5", self.driver.page_source.encode('utf-8')).hexdigest()  # nosec
         if self.last_page_hash != hash:
             self.save_response()
 
@@ -806,7 +801,7 @@ class SeleniumBrowser(object):
         self.driver.execute(Command.CLEAR_LOCAL_STORAGE)
 
 
-class SubSeleniumMixin(object):
+class SubSeleniumMixin:
     """Mixin to have a Selenium browser for performing login."""
 
     SELENIUM_BROWSER = None
