@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2016 Romain Bignon
 #
 # This file is part of woob.
@@ -20,20 +18,24 @@
 
 from binascii import crc32
 import re
+from typing import Optional, Iterator, List
 
+from woob.capabilities.account import CapCredentialsCheck
 from woob.capabilities.base import (
-    BaseObject, Field, StringField, DecimalField, IntField,
-    UserError, Currency, NotAvailable, EnumField, Enum,
-    empty, find_object
+    BaseObject, Capability, Field, StringField, DecimalField, IntField,
+    BoolField, UserError, Currency, NotAvailable, EnumField, Enum,
+    empty, find_object,
 )
 from woob.capabilities.date import DateField
 from woob.capabilities.collection import CapCollection
-from woob.tools.compat import unicode
+from woob.exceptions import BrowserIncorrectPassword
 
 
 __all__ = [
     'CapBank', 'BaseAccount', 'Account', 'Loan', 'Transaction', 'AccountNotFound',
-    'AccountType', 'AccountOwnership',
+    'AccountType', 'AccountOwnership', 'Balance', 'AccountSchemeName', 'TransactionCounterparty',
+    'AccountOwnerProfile', 'IdentityParty', 'AccountParty', 'AccountIdentification',
+    'PartyRole', 'CapAccountCheck',
 ]
 
 
@@ -46,8 +48,8 @@ class AccountNotFound(ObjectNotFound):
     Raised when an account is not found.
     """
 
-    def __init__(self, msg='Account not found'):
-        super(AccountNotFound, self).__init__(msg)
+    def __init__(self, msg: Optional[str] = 'Account not found'):
+        super().__init__(msg)
 
 
 class BaseAccount(BaseObject, Currency):
@@ -59,15 +61,15 @@ class BaseAccount(BaseObject, Currency):
     currency =       StringField('Currency', default=None)
     bank_name =      StringField('Bank Name', mandatory=False)
 
-    def __init__(self, id='0', url=None):
-        super(BaseAccount, self).__init__(id, url)
+    def __init__(self, id: Optional[str] = '0', url: Optional[str] = None):
+        super().__init__(id, url)
 
     @property
-    def currency_text(self):
+    def currency_text(self) -> str:
         return Currency.currency2txt(self.currency)
 
     @property
-    def ban(self):
+    def ban(self) -> str:
         """ Bank Account Number part of IBAN"""
         if not self.iban:
             return NotAvailable
@@ -118,32 +120,134 @@ class AccountType(Enum):
     "Pension plan PER"
     REAL_ESTATE = 21
     "Real estate investment such as SCPI, OPCI, SCI"
+    CROWDLENDING = 22
+    "Crowdlending accounts"
 
 
-class AccountOwnerType(object):
+class AccountOwnerType:
     """
     Specifies the usage of the account
     """
-    PRIVATE = u'PRIV'
+    PRIVATE = 'PRIV'
     """private personal account"""
-    ORGANIZATION = u'ORGA'
+    ORGANIZATION = 'ORGA'
     """professional account"""
-    ASSOCIATION = u'ASSO'
+    ASSOCIATION = 'ASSO'
     """association account"""
 
 
-class AccountOwnership(object):
+class AccountOwnership:
     """
     Relationship between the credentials owner (PSU) and the account
     """
-    OWNER = u'owner'
+    OWNER = 'owner'
     """The PSU is the account owner"""
-    CO_OWNER = u'co-owner'
+    CO_OWNER = 'co-owner'
     """The PSU is the account co-owner"""
-    ATTORNEY = u'attorney'
+    ATTORNEY = 'attorney'
     """The PSU is the account attorney"""
 
 
+class AccountSchemeName(Enum):
+    IBAN = 'iban'
+    """IBAN as defined in ISO 13616"""
+
+    BBAN = 'bban'
+    """Basic Bank Account Number, represents a country-specific bank account number"""
+
+    SORT_CODE_ACCOUNT_NUMBER = 'sort_code_account_number'
+    """Account Identification Number sometimes employed instead of IBAN (e.g.: in UK)"""
+
+    CPAN = 'cpan'
+    """Card PAN (masked or plain)"""
+
+
+class TransactionCounterparty(BaseObject):
+    label = StringField('Name of the other stakeholder (Creditor or debtor)', default=None)
+    account_scheme_name = EnumField('Type of account Scheme', AccountSchemeName, default=None)
+    account_identification = StringField('ID of the account', default=None)
+    debtor = BoolField('Type of the counterparty (debtor/creditor/null)', default=None)
+
+    def __repr__(self):
+        return f'<label={self.label} debtor={self.debtor} account_scheme_name={self.account_scheme_name} account_identification={self.account_identification}>'
+
+
+class AccountOwnerProfile(TransactionCounterparty):
+    """
+    AccountOwnerProfile contains the information related to the account and its owner:
+    - account_scheme_name: scheme type of the account (IBAN or sort code + account number)
+    - account_identification: value that identifies the account and is related to the scheme type
+    - holder_name: full name of the account owner
+    """
+    holder_name = StringField('Name of the account holder', default=None)
+
+    def __repr__(self):
+        return f'<AccountOwnerProfile account_scheme_name={self.account_scheme_name} account_identification={self.account_identification} holder_name={self.holder_name}>'
+
+
+class PartyRole(Enum):
+    UNKNOWN = 'unknown'
+    HOLDER = 'holder'
+    CO_HOLDER = 'co_holder'
+    ATTORNEY = 'attorney'
+    CUSTODIAN_FOR_MINOR = 'custodian_for_minor'
+    LEGAL_GUARDIAN = 'legal_guardian'
+    NOMINEE_BENEFICIARY = 'nominee_beneficiary'
+    SUCCESSOR_ON_DEATH = 'successor_on_death'
+    TRUSTEE = 'trustee'
+
+
+class AccountIdentification(BaseObject):
+    """
+    Defines the identification of a account:
+    - scheme_name: Name of the account scheme type
+    - identification: ID of the account
+    """
+    scheme_name = EnumField('Name of the account scheme type', AccountSchemeName, default=None)
+    identification = StringField('ID of the account', default=None)
+
+    def __repr__(self):
+        return f'<AccountIdentification scheme_name={self.scheme_name} identification={self.identification}>'
+
+
+class IdentityParty(BaseObject):
+    """
+    Defines the identity of a party:
+    - full_name: Full name of the party
+    - role: Role of the party
+    - is_user: Defines the link between the party and the connected PSU
+    """
+    ROLE_UNKNOWN = PartyRole.UNKNOWN
+    ROLE_HOLDER = PartyRole.HOLDER
+    ROLE_CO_HOLDER = PartyRole.CO_HOLDER
+    ROLE_ATTORNEY = PartyRole.ATTORNEY
+    ROLE_CUSTODIAN_FOR_MINOR = PartyRole.CUSTODIAN_FOR_MINOR
+    ROLE_LEGAL_GUARDIAN = PartyRole.LEGAL_GUARDIAN
+    ROLE_NOMINEE_BENEFICIARY = PartyRole.NOMINEE_BENEFICIARY
+    ROLE_SUCCESSOR_ON_DEATH = PartyRole.SUCCESSOR_ON_DEATH
+    ROLE_TRUSTEE = PartyRole.TRUSTEE
+
+    full_name = StringField('Full name of the party.', default=None)
+    is_user = BoolField('Is the party the connected PSU?', default=None)
+    role = EnumField('Role of the party.', PartyRole, default=ROLE_UNKNOWN)
+
+    def __repr__(self):
+        return f'<IdentityParty full_name={self.full_name} role={self.role} is_user={self.is_user}>'
+
+
+class AccountParty(BaseObject):
+    """
+    Defines all the information related to an account party:
+    - identities: list of IdentityParty elements
+    - identifications : list of AccountIdentification elements
+    """
+    identities = Field('Identities of the account party', list, default=[])
+    identifications = Field('Identification information of the account', list, default=[])
+
+    def __repr__(self):
+        return f'<AccountParty identities={self.identities} identifications={self.identifications}>'
+
+
 class Account(BaseAccount):
     """
     Bank account.
@@ -170,6 +274,7 @@ class Account(BaseAccount):
     TYPE_REVOLVING_CREDIT = AccountType.REVOLVING_CREDIT
     TYPE_PER              = AccountType.PER
     TYPE_REAL_ESTATE      = AccountType.REAL_ESTATE
+    TYPE_CROWDLENDING     = AccountType.CROWDLENDING
 
     type =      EnumField('Type of account', AccountType, default=TYPE_UNKNOWN)
     owner_type = StringField('Usage of account')  # cf AccountOwnerType class
@@ -201,6 +306,12 @@ class Account(BaseAccount):
 
     opening_date = DateField('Date when the account contract was created on the bank')
 
+    all_balances = Field('List of balances', list, default=[])
+
+    profile = Field('Profile associated to the account owner', AccountOwnerProfile)
+
+    party = Field('Party associated to the account', AccountParty, default=None, mandatory=False)
+
     def __repr__(self):
         return "<%s id=%r label=%r>" % (type(self).__name__, self.id, self.label)
 
@@ -214,6 +325,30 @@ class Account(BaseAccount):
         self.valuation_diff_ratio = value
 
 
+class BalanceType(Enum):
+    CLOSING = 1
+    """Current balance of the account"""
+    PENDING = 2
+    """Forecast balance of the account"""
+
+
+class Balance(BaseObject):
+    """
+    Object made to receive balance on one Account
+    """
+
+    amount = DecimalField('Amount on this balance')
+    type = EnumField('Type of balance', BalanceType)
+    currency = StringField('Currency')
+    reference_date = DateField('date of the balance')
+    last_update = DateField('Last time balance was updated')
+    credit_included = BoolField('If factoring is included in balance', default=False)
+    label = StringField('Bank name of the balance')
+
+    def __repr__(self):
+        return f'<{type(self).__name__} label={self.label} amount={self.amount} type={self.type} credit_included={self.credit_included}>'
+
+
 class Loan(Account):
     """
     Account type dedicated to loans and credits.
@@ -227,6 +362,9 @@ class Loan(Account):
     available_amount = DecimalField('Amount available') # only makes sense for revolving credit
     used_amount = DecimalField('Amount already used') # only makes sense for revolving credit
 
+    insurance_amount = DecimalField("Amount of the loan's insurance")
+    insurance_rate = DecimalField("Rate of the loan's insurance")
+
     subscription_date = DateField('Date of subscription of the loan')
     maturity_date = DateField('Estimated end date of the loan')
     duration = IntField('Duration of the loan given in months')
@@ -256,6 +394,7 @@ class TransactionType(Enum):
     CASH_DEPOSIT  = 10
     CARD_SUMMARY  = 11
     DEFERRED_CARD = 12
+    INSTANT       = 13
 
 
 class Transaction(BaseObject):
@@ -275,6 +414,7 @@ class Transaction(BaseObject):
     TYPE_CASH_DEPOSIT  = TransactionType.CASH_DEPOSIT
     TYPE_CARD_SUMMARY  = TransactionType.CARD_SUMMARY
     TYPE_DEFERRED_CARD = TransactionType.DEFERRED_CARD
+    TYPE_INSTANT       = TransactionType.INSTANT
 
     date =      DateField('Debit date on the bank statement')
     rdate =     DateField('Real date, when the payment has been made; usually extracted from the label or from credit card info')
@@ -302,17 +442,19 @@ class Transaction(BaseObject):
     # Financial arbitrations
     investments =       Field('List of investments related to the transaction', list, default=[])
 
+    counterparty = Field('Counterparty of transaction', TransactionCounterparty)
+
     def __repr__(self):
         return "<Transaction date=%r label=%r amount=%r>" % (self.date, self.label, self.amount)
 
-    def unique_id(self, seen=None, account_id=None):
+    def unique_id(self, seen: Optional[set] = None, account_id: Optional[str] = None) -> str:
         """
         Get an unique ID for the transaction based on date, amount and raw.
 
-        :param seen: if given, the method uses this dictionary as a cache to
+        :param seen: if given, the method uses this set as a cache to
                      prevent several transactions with the same values to have the same
                      unique ID.
-        :type seen: :class:`dict`
+        :type seen: :class:`set`
         :param account_id: if given, add the account ID in data used to create
                            the unique ID. Can be useful if you want your ID to be unique across
                            several accounts.
@@ -320,8 +462,8 @@ class Transaction(BaseObject):
         :returns: an unique ID encoded in 8 length hexadecimal string (for example ``'a64e1bc9'``)
         :rtype: :class:`str`
         """
-        crc = crc32(unicode(self.date).encode('utf-8'))
-        crc = crc32(unicode(self.amount).encode('utf-8'), crc)
+        crc = crc32(str(self.date).encode('utf-8'))
+        crc = crc32(str(self.amount).encode('utf-8'), crc)
         if not empty(self.raw):
             label = self.raw
         else:
@@ -330,7 +472,7 @@ class Transaction(BaseObject):
         crc = crc32(re.sub('[ ]+', ' ', label).encode("utf-8"), crc)
 
         if account_id is not None:
-            crc = crc32(unicode(account_id).encode('utf-8'), crc)
+            crc = crc32(str(account_id).encode('utf-8'), crc)
 
         if seen is not None:
             while crc in seen:
@@ -341,12 +483,31 @@ class Transaction(BaseObject):
         return "%08x" % (crc & 0xffffffff)
 
 
-class CapBank(CapCollection):
+class CapBank(CapCollection, CapCredentialsCheck):
     """
     Capability of bank websites to see accounts and transactions.
     """
 
-    def iter_resources(self, objs, split_path):
+    def check_credentials(self) -> bool:
+        """
+        Check that the given credentials are correct by trying to login.
+
+        The default implementation of this method check if the class using this capability
+        has a browser, execute its do_login if it has one and then see if no error pertaining to the creds is raised.
+        If any other unexpected error occurs, we don't know whether the creds are correct or not.
+        """
+        # TODO move this in a specific capability
+        if getattr(self, 'BROWSER', None) is None:
+            raise NotImplementedError()
+
+        try:
+            self.browser.do_login()
+        except BrowserIncorrectPassword:
+            return False
+
+        return True
+
+    def iter_resources(self, objs: List[BaseObject], split_path: List[str]) -> Iterator[BaseObject]:
         """
         Iter resources.
 
@@ -364,7 +525,7 @@ class CapBank(CapCollection):
 
             return self.iter_accounts()
 
-    def iter_accounts(self):
+    def iter_accounts(self) -> Iterator[Account]:
         """
         Iter accounts.
 
@@ -372,7 +533,7 @@ class CapBank(CapCollection):
         """
         raise NotImplementedError()
 
-    def get_account(self, id):
+    def get_account(self, id: str) -> Account:
         """
         Get an account from its ID.
 
@@ -383,7 +544,7 @@ class CapBank(CapCollection):
         """
         return find_object(self.iter_accounts(), id=id, error=AccountNotFound)
 
-    def iter_history(self, account):
+    def iter_history(self, account: Account) -> Iterator[Account]:
         """
         Iter history of transactions on a specific account.
 
@@ -394,7 +555,7 @@ class CapBank(CapCollection):
         """
         raise NotImplementedError()
 
-    def iter_coming(self, account):
+    def iter_coming(self, account: Account) -> Iterator[Transaction]:
         """
         Iter coming transactions on a specific account.
 
@@ -404,3 +565,19 @@ class CapBank(CapCollection):
         :raises: :class:`AccountNotFound`
         """
         raise NotImplementedError()
+
+
+class CapAccountCheck(Capability):
+    """
+    Capability to get accounts parties information.
+
+    The expected structure is the following:
+        - AccountParty object
+            * identities (list of IdentityParty elements):
+                * full name
+                * role
+                * is_user
+            * identifications (list of type AccountIdentification elements):
+                * scheme name
+                * identification
+    """
