@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2016 Romain Bignon
 #
 # This file is part of woob.
@@ -18,8 +16,11 @@
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
 
-from datetime import date, datetime
+import enum
 import re
+from datetime import date, datetime
+from typing import Iterable
+
 from unidecode import unidecode
 
 from woob.capabilities.base import (
@@ -71,7 +72,8 @@ class TransferError(UserError):
 
     def __init__(self, description=None, message=None):
         """
-        :param message: error message from the bank, if any
+        :param description: technical description of the error
+        :param message: error message from the bank, if any, to display to the user
         """
 
         super(TransferError, self).__init__(message or description)
@@ -152,7 +154,8 @@ class AddRecipientError(UserError):
 
     def __init__(self, description=None, message=None):
         """
-        :param message: error message from the bank, if any
+        :param description: technical description of the error
+        :param message: error message from the bank, if any, to display to the user
         """
 
         super(AddRecipientError, self).__init__(message or description)
@@ -212,10 +215,25 @@ class AddRecipientStep(BrowserQuestion):
         self.recipient = recipient
 
 
-class BeneficiaryType(object):
-    RECIPIENT =          'recipient'
-    IBAN =               'iban'
-    PHONE_NUMBER =       'phone_number'
+class BeneficiaryType(Enum):
+    """
+    Type of the Transfer.beneficiary_number property.
+    """
+
+    IBAN = 'iban'
+    """beneficiary number is an IBAN as defined in ISO 13616"""
+
+    SORT_CODE_ACCOUNT_NUMBER = 'sort_code_account_number'
+    """account number is a national UK/Ireland number including sortcode"""
+
+    PHONE_NUMBER = 'phone_number'
+    """beneficiary number is an E.164 encoded phone number"""
+
+    RECIPIENT = 'recipient'
+    """
+    beneficiary number is a beneficiary identifier as returned by
+    the iter_transfer_recipients method.
+    """
 
 
 class TransferStatus(Enum):
@@ -227,22 +245,36 @@ class TransferStatus(Enum):
     SCHEDULED = 'scheduled'
     """Transfer to be executed later"""
 
+    ACTIVE = 'active'
+    """Periodic transfer is still active"""
+
     DONE = 'done'
     """Transfer was executed"""
 
     CANCELLED = 'cancelled'
     """Transfer was cancelled by the bank or by the user"""
 
+    ACCEPTED_NO_BANK_STATUS = 'accepted_no_bank_status'
+    """Transfer was sent to the bank but we will not get more information
+    after that. This is used for banks that do not give us final states after pending."""
+
 
 class TransferFrequency(Enum):
     UNKNOWN = 'unknown'
+    DAILY = 'daily'
     WEEKLY = 'weekly'
+    TWOWEEKLY = 'two-weekly'  # every two weeks, not 2 times per week
     MONTHLY = 'monthly'
-    BIMONTHLY = 'bimonthly'
+    TWOMONTHLY  = 'two-monthly'  # every two months, not 2 times per month
     QUARTERLY = 'quarterly'
-    BIANNUAL = 'biannual'
+    FOURMONTHLY = 'four-monthly'  # every four months, not 4 times per month
+    SEMIANNUALLY = 'semiannually'
     YEARLY = 'yearly'
 
+    # (deprecated)
+    BIMONTHLY = 'bimonthly'  # use TWOWEEKLY instead
+    BIANNUAL = 'biannual'  # use SEMIANNUALLY instead
+
 
 class TransferDateType(Enum):
     FIRST_OPEN_DAY = 'first_open_day'
@@ -297,6 +329,9 @@ class Transfer(BaseObject, Currency):
 
     cancelled_exception = Field('Transfer cancelled reason', TransferError)
 
+    # End to end ID given by the client
+    reference_id = StringField('End to end ID given by client')
+
 
 class EmitterNumberType(Enum):
     UNKNOWN = 'unknown'
@@ -313,6 +348,32 @@ class Emitter(BaseAccount):
     balance = DecimalField('Balance of emitter account')
 
 
+class DebtorAccountRequirement(Enum):
+    MANDATORY = 'mandatory'
+    """Debtor account is needed to initiate a payment"""
+
+    OPTIONAL = 'optional'
+    """Debtor account is optional (may change module behaviour)"""
+
+    NOT_USED = 'not_used'
+    """Debtor account must not be given"""
+
+
+class Platform(str, enum.Enum):
+    """Mobile platform on which the webview can be run.
+
+    For instance, this enumeration can be used to represent systems on which
+    the authorization link can be catched by a native application instead of
+    the browser.
+    """
+
+    ANDROID = 'android'
+    """Android based platforms."""
+
+    IOS = 'ios'
+    """Apple's iOS platform."""
+
+
 class CapTransfer(Capability):
     can_do_transfer_to_untrusted_beneficiary = False
     """
@@ -322,7 +383,7 @@ class CapTransfer(Capability):
     in `iter_transfer_recipients` like for PSD2 modules
     """
 
-    can_do_transfer_without_emitter = False
+    can_do_transfer_without_emitter = True
     """
     The module can do transfer without giving the emitter, for example:
     when there is only, and will be only, one account like wallet
@@ -330,8 +391,52 @@ class CapTransfer(Capability):
     and the emitter is chosen afterwards like for PSD2 modules
     """
 
+    can_do_transfer_cancellation = False
+
+    sca_required_for_transfer_cancellation = False
+    """
+    The default behavior is that we don't need to validate a payment cancellation through a SCA.
+    If a SCA is required after sending a transfer cancellation request to validate it, the module
+    should set this to True.
+    """
+
     accepted_beneficiary_types = (BeneficiaryType.RECIPIENT, )
     accepted_execution_date_types = (TransferDateType.FIRST_OPEN_DAY, TransferDateType.DEFERRED)
+    accepted_execution_frequencies = set(TransferFrequency) - set([TransferFrequency.UNKNOWN])
+    maximum_number_of_instructions = 1
+    transfer_with_debtor_account = DebtorAccountRequirement.NOT_USED
+
+    # Indicate that we may not know if the payment is done or rejected, the information
+    # is provided by date type because this behaviour is generally dependent on the type
+    # of payment. An empty list means that the transfer will never get the status ACCEPTED_NO_BANK_STATUS
+    partial_transfer_status_tracking = ()
+
+    is_app_to_app_used_for_transfer = {
+        Platform.ANDROID: None,
+        Platform.IOS: None,
+    }  # type: dict[Platform, bool | None]
+    """
+    Is an App2App flow used for the payment if the PSU has the bank's app installed.
+    None means unknown
+    """
+
+    bank_provides_payer_account = None  # type: bool | None
+    """
+    Once the payment is initiated, does the bank return the payer's account identifier?
+    None means unknown
+    """
+
+    bank_provides_payer_label = None  # type: bool | None
+    """
+    Once the payment is initiated, does the bank return the payer's label?
+    None means unknown
+    """
+
+    transfer_date_types_where_trusted_beneficiary_required = set()  # type: Iterable[TransferDateType]
+    """
+    Set of `TransferDateType` where the beneficiary must be trusted or registered on the payer's banking service.
+    If `iter_transfer_recipients` is implemented, such beneficiaries may be found.
+    """
 
     def iter_transfer_recipients(self, account):
         """
@@ -376,6 +481,35 @@ class CapTransfer(Capability):
         """
         return self.get_transfer(transfer.id)
 
+    def confirm_transfer_cancellation(self, transfer, **params):
+        """
+        Confirm transfer cancellation after a redirect flow.
+
+        :param :class:`Transfer`
+        :rtype: :class:`Transfer`
+        :raises: :class:`AssertionError`: If the payment is not actually cancelled after the whole process
+        """
+        transfer = self.optional_confirm_transfer_cancellation(transfer, **params)
+        # Check that the transfer has been successfully cancelled.
+        if transfer.status != TransferStatus.CANCELLED:
+            raise AssertionError('Transfer is not cancelled after cancellation request.')
+        return transfer
+
+    def optional_confirm_transfer_cancellation(self, transfer, **params):
+        """Proceed with the actual cancellation confirmation.
+
+        This method MUST NOT be called by any external caller. Said caller
+        should actually call confirm_transfer_cancellation which may call the
+        current method if it sees fit.
+
+        The default implementation does not run an explicit confirmation step,
+        it only fetches the up-to-date transfer.
+
+        Modules requiring an explicit cancellation confirmation should
+        overwrite this method, returning the up-to-date transfer at the end.
+        """
+        return self.get_transfer(transfer.id)
+
     def transfer(self, transfer, **params):
         """
         Do a transfer from an account to a recipient.
@@ -389,19 +523,62 @@ class CapTransfer(Capability):
             BeneficiaryType.RECIPIENT: ('id', 'beneficiary_number', 'beneficiary_label',),
             BeneficiaryType.IBAN: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
             BeneficiaryType.PHONE_NUMBER: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
+            BeneficiaryType.SORT_CODE_ACCOUNT_NUMBER: ('id', 'recipient_id', 'recipient_iban', 'recipient_label',),
         }
 
-        if not transfer.amount or transfer.amount <= 0:
-            raise TransferInvalidAmount('amount must be strictly positive')
+        if hasattr(transfer, 'instructions'):
+            instructions = transfer.instructions
+        else:
+            instructions = [transfer]
+        nb_instructions = len(instructions)
 
+        for instr in instructions:
+            if not instr.amount or instr.amount <= 0:
+                raise TransferInvalidAmount('amount must be strictly positive')
+
+        instructions = sorted(
+            instructions,
+            key=lambda x: (x.reference_id, x.beneficiary_number, x.recipient_iban, x.amount, x.exec_date)
+        )
+
+        # Initiate the transfer
         t = self.init_transfer(transfer, **params)
-        for key, value in t.iter_fields():
-            if hasattr(transfer, key) and (key not in transfer_not_check_fields[transfer.beneficiary_type]):
-                transfer_val = getattr(transfer, key)
-                if hasattr(self, 'transfer_check_%s' % key):
-                    assert getattr(self, 'transfer_check_%s' % key)(transfer_val, value), '%s changed during transfer processing (from "%s" to "%s")' % (key, transfer_val, value)
-                else:
-                    assert transfer_val == value or empty(transfer_val), '%s changed during transfer processing (from "%s" to "%s")' % (key, transfer_val, value)
+
+        # Verify the created transfer before execution
+        if hasattr(t, 'instructions'):
+            new_instructions = t.instructions
+        else:
+            new_instructions = [t]
+        nb_new_instructions = len(new_instructions)
+        new_instructions = sorted(
+            new_instructions,
+            key=lambda x: (x.reference_id, x.beneficiary_number, x.recipient_iban, x.amount, x.exec_date)
+        )
+
+        assert nb_instructions == nb_new_instructions, (
+            'Number of instructions changed during transfer processing (from "%s" to "%s")' % (nb_instructions, nb_new_instructions)
+        )
+
+        changed_msg_template = '%s changed during transfer processing (from "%s" to "%s") for instruction %s'
+        for orig_instr, new_instr in zip(instructions, new_instructions):
+            ignored_keys = transfer_not_check_fields[orig_instr.beneficiary_type]
+            for key, value in new_instr.iter_fields():
+                if key in ignored_keys:
+                    continue
+                try:
+                    transfer_val = getattr(orig_instr, key)
+                except AttributeError:
+                    continue
+                transfer_check_fn = getattr(self, 'transfer_check_%s' % key, None)
+                if transfer_check_fn:
+                    if not transfer_check_fn(transfer_val, value):
+                        raise AssertionError(changed_msg_template % (
+                            key, transfer_val, value, orig_instr.reference_id or ''
+                        ))
+                elif transfer_val != value and not empty(transfer_val):
+                    raise AssertionError(changed_msg_template % (
+                        key, transfer_val, value, orig_instr.reference_id or ''
+                    ))
         return self.execute_transfer(t, **params)
 
     def transfer_check_label(self, old, new):
@@ -438,8 +615,36 @@ class CapTransfer(Capability):
         """
         raise NotImplementedError()
 
+    def cancel_transfer(self, transfer, **params):
+        """
+        Ask for the cancellation of a transfer.
+
+        This function is exposed as part of Woob API and should not be overriden
+        by children modules.
+
+        :param transfer: the transfer that should be cancelled
+        :type transfer: :class:`Transfer`
+        :rtype: :class:`Transfer`
+        """
+        return self.do_transfer_cancellation(transfer, **params)
+
+    def do_transfer_cancellation(self, transfer, **params):
+        """
+        Send a cancellation request for the given transfer.
+
+        This function should be implemented by the children modules.
+
+        :param transfer: the transfer that should be cancelled
+        :type transfer: :class:`Transfer`
+        :rtype: :class:`Transfer`
+        """
+        raise NotImplementedError()
+
 
 class CapBankTransfer(CapBank, CapTransfer):
+    can_do_transfer_without_emitter = False
+    transfer_with_debtor_account = DebtorAccountRequirement.MANDATORY
+
     def account_to_emitter(self, account):
         if isinstance(account, Account):
             account = account.id
