@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2013 Christophe Benz, Romain Bignon, Julien Hebert
 #
 # This file is part of woob.
@@ -17,14 +15,12 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
-from collections import OrderedDict
+from collections import OrderedDict, deque
 import warnings
 import re
 from decimal import Decimal
 from copy import deepcopy, copy
-import sys
 
-from woob.tools.compat import unicode, long, with_metaclass, StrConv
 from woob.tools.misc import to_unicode
 
 
@@ -46,9 +42,6 @@ class EnumMeta(type):
     def __init__(cls, name, bases, attrs, *args, **kwargs):
         super(EnumMeta, cls).__init__(name, bases, attrs, *args, **kwargs)
         attrs = [(k, v) for k, v in attrs.items() if not callable(v) and not k.startswith('__')]
-        if sys.version_info.major < 3:
-            # can't have original declaration order, at least sort by value
-            attrs.sort(key=lambda kv: kv[1])
         cls.__members__ = OrderedDict(attrs)
 
     def __setattr__(cls, name, value):
@@ -88,7 +81,7 @@ class EnumMeta(type):
         return cls.__members__[k]
 
 
-class Enum(with_metaclass(EnumMeta, object)):
+class Enum(metaclass=EnumMeta):
     pass
 
 
@@ -118,6 +111,88 @@ def find_object(mylist, error=None, **kwargs):
     return None
 
 
+def find_object_any_match(objects, key_value_pairs, error=None, with_priority=True, ignore_empty=True):
+    """
+    Tool method that returns an object that match any parameter given in key_value_pairs.
+
+    Compared with find_object, this method does not need all parameters to be matching.
+    If no priority is set, the method will return the first successful match. With several key-values set,
+    this may not be the object instance with the most key matches.
+
+    :param objects: The list or iterator of objects that can match
+    :type objectlist: list
+    :param key_value_pairs: The key-values that the object should have to match. This is given as a list
+        of pairs, where each pair represent a key-value, the first item being the key, and the second is the value.
+    :type key_value_pairs: list
+    :param error: The error to raise in case no object matches.
+    :type error: Exception or None
+    :param with_priority: If True, the key-value pairs order decides the priority of the matching.
+        With priority enabled, an object matching with a more important key will be matched over another
+        object matching with a less important keys. This means that all objects will be iterated in order
+        to assure this priority, unless the most preferred key matches, in which case the iteration stops early.
+        This feature is enabled by default. See the example sections for more details.
+    :type with_priority: bool
+    :param ignore_empty: If True, empty key-values (None, EmptyType) will not be matched. True by default.
+    :type ignore_empty: bool
+
+    :return: The matching object if any
+
+    Examples:
+
+    >>> class Point:
+    ...     def __init__(self, x, y):
+    ...         self.x = x
+    ...         self.y = y
+    ...     def __repr__(self):
+    ...         return f'Point({self.x},{self.y})'
+    >>> points = (Point(0,0), Point(1, 2), Point(None, 3))
+    >>> find_object_any_match(points, (('x', 0), ('y', 1))) # Will match on x with first point
+    Point(0,0)
+    >>> find_object_any_match(points, (('x', 2), ('y', 2))) # Will match on y with second point
+    Point(1,2)
+    >>> find_object_any_match(points, (('x', 1), ('y', 0))) # Will match on x with second point because priority
+    Point(1,2)
+    >>> find_object_any_match(points, (('x', 1), ('y', 0)), with_priority=False) # Will return first successful match
+    Point(0,0)
+    >>> find_object_any_match(points, (('x', None),)) # second condition (None) is ignored by default
+    None
+    >>> find_object_any_match(points, (('x', None),), ignore_empty=False) # will match last point
+    Point(None,3)
+    >>> find_object_any_match(points, (('x', -1),), error=ValueError) # Will raise ValueError
+    ValueError:
+    """
+
+    # Will be used if with_priority is true
+    found = None
+    found_index = None
+
+    for o in objects:
+        for index, (key, value) in enumerate(key_value_pairs):
+            # There is no point trying to match as it will be not priorized
+            if found_index and found_index <= index:
+                break
+
+            o_value = getattr(o, key)
+            if (
+                ignore_empty and (empty(value) or empty(o_value))
+                or o_value != value
+            ):
+                continue
+
+            # First key is a special case with priority, we can return early
+            # if we find a match on the first key
+            # Also we only keep more priorized matches
+            if not with_priority or index == 0:
+                return o
+
+            found = o
+            found_index = index
+
+    if not found and error is not None:
+        raise error()
+    return found
+
+
 def strict_find_object(mylist, error=None, **kwargs):
     """
     Tools to return an object with the matching parameters in kwargs.
@@ -148,7 +223,7 @@ class FieldNotFound(Exception):
     """
 
     def __init__(self, obj, field):
-        super(FieldNotFound, self).__init__(u'Field "%s" not found for object %s' % (field, obj))
+        super(FieldNotFound, self).__init__('Field "%s" not found for object %s' % (field, obj))
 
 
 class ConversionWarning(UserWarning):
@@ -166,7 +241,7 @@ class AttributeCreationWarning(UserWarning):
     """
 
 
-class EmptyType(object):
+class EmptyType:
     """
     Parent class for NotAvailableType, NotLoadedType and FetchErrorType.
     """
@@ -194,8 +269,8 @@ class NotAvailableType(EmptyType):
     def __repr__(self):
         return 'NotAvailable'
 
-    def __unicode__(self):
-        return u'Not available'
+    def __str__(self):
+        return 'Not available'
 
 
 NotAvailable = NotAvailableType()
@@ -212,8 +287,8 @@ class NotLoadedType(EmptyType):
     def __repr__(self):
         return 'NotLoaded'
 
-    def __unicode__(self):
-        return u'Not loaded'
+    def __str__(self):
+        return 'Not loaded'
 
 
 NotLoaded = NotLoadedType()
@@ -227,14 +302,14 @@ class FetchErrorType(EmptyType):
     def __repr__(self):
         return 'FetchError'
 
-    def __unicode__(self):
-        return u'Not mandatory'
+    def __str__(self):
+        return 'Not mandatory'
 
 
 FetchError = FetchErrorType()
 
 
-class Capability(object):
+class Capability:
     """
     This is the base class for all capabilities.
 
@@ -245,7 +320,7 @@ class Capability(object):
     """
 
 
-class Field(object):
+class Field:
     """
     Field of a :class:`BaseObject` class.
 
@@ -280,11 +355,11 @@ class Field(object):
 
 class IntField(Field):
     """
-    A field which accepts only :class:`int` and :class:`long` types.
+    A field which accepts only :class:`int` types.
     """
 
     def __init__(self, doc, **kwargs):
-        super(IntField, self).__init__(doc, int, long, **kwargs)
+        super(IntField, self).__init__(doc, int, **kwargs)
 
     def convert(self, value):
         return int(value)
@@ -330,11 +405,11 @@ class FloatField(Field):
 
 class StringField(Field):
     """
-    A field which accepts only :class:`unicode` strings.
+    A field which accepts only :class:`str` strings.
     """
 
     def __init__(self, doc, **kwargs):
-        super(StringField, self).__init__(doc, unicode, **kwargs)
+        super(StringField, self).__init__(doc, str, **kwargs)
 
     def convert(self, value):
         return to_unicode(value)
@@ -349,7 +424,7 @@ class BytesField(Field):
         super(BytesField, self).__init__(doc, bytes, **kwargs)
 
     def convert(self, value):
-        if isinstance(value, unicode):
+        if isinstance(value, str):
             value = value.encode('utf-8')
         return bytes(value)
 
@@ -381,7 +456,7 @@ class _BaseObjectMeta(type):
 
         if new_class.__doc__ is None:
             new_class.__doc__ = ''
-        for name, field in fields:
+        for name, field in new_class._fields.items():
             doc = '(%s) %s' % (', '.join([':class:`%s`' % v.__name__ if isinstance(v, type) else v for v in field.types]), field.doc)
             if field.value is not NotLoaded:
                 doc += ' (default: %s)' % field.value
@@ -389,7 +464,7 @@ class _BaseObjectMeta(type):
         return new_class
 
 
-class BaseObject(with_metaclass(_BaseObjectMeta, StrConv, object)):
+class BaseObject(metaclass=_BaseObjectMeta):
     """
     This is the base class for a capability object.
 
@@ -406,9 +481,9 @@ class BaseObject(with_metaclass(_BaseObjectMeta, StrConv, object)):
             " Transfer from an account to a recipient.  "
 
             amount =    DecimalField('Amount to transfer')
-            date =      Field('Date of transfer', basestring, date, datetime)
-            origin =    Field('Origin of transfer', int, long, basestring)
-            recipient = Field('Recipient', int, long, basestring)
+            date =      Field('Date of transfer', str, date, datetime)
+            origin =    Field('Origin of transfer', int, str)
+            recipient = Field('Recipient', int, str)
 
     The docstring is mandatory.
     """
@@ -418,8 +493,8 @@ class BaseObject(with_metaclass(_BaseObjectMeta, StrConv, object)):
     url = StringField('url')
     _fields = None
 
-    def __init__(self, id=u'', url=NotLoaded, backend=None):
-        self.id = to_unicode(id) if id is not None else u''
+    def __init__(self, id='', url=NotLoaded, backend=None):
+        self.id = to_unicode(id) if id is not None else ''
         self.backend = backend
         self._fields = deepcopy(self._fields)
         self.__setattr__('url', url)
@@ -507,41 +582,20 @@ class BaseObject(with_metaclass(_BaseObjectMeta, StrConv, object)):
                 try:
                     # Try to convert value to the wanted one.
                     nvalue = attr.convert(value)
+                except (TypeError, ValueError, ArithmeticError):
+                    # error during conversion, it will probably not
+                    # match the wanted following types, so we'll
+                    # raise ValueError.
+                    pass
+                else:
                     # If the value was converted
                     if nvalue is not value:
                         warnings.warn('Value %s was converted from %s to %s' %
                                       (name, type(value), type(nvalue)),
                                       ConversionWarning, stacklevel=2)
                     value = nvalue
-                except Exception:
-                    # error during conversion, it will probably not
-                    # match the wanted following types, so we'll
-                    # raise ValueError.
-                    pass
-            from collections import deque
-            actual_types = ()
-            for v in attr.types:
-                if isinstance(v, str):
-                    # the following is a (almost) copy/paste from
-                    # https://stackoverflow.com/questions/11775460/lexical-cast-from-string-to-type
-                    q = deque([object])
-                    while q:
-                        t = q.popleft()
-                        if t.__name__ == v:
-                            actual_types += (t,)
-                        else:
-                            try:
-                                # keep looking!
-                                q.extend(t.__subclasses__())
-                            except TypeError:
-                                # type.__subclasses__ needs an argument for
-                                # whatever reason.
-                                if t is type:
-                                    continue
-                                else:
-                                    raise
-                else:
-                    actual_types += (v,)
+
+            actual_types = _resolve_types(attr.types)
 
             if not isinstance(value, actual_types) and not empty(value):
                 raise ValueError(
@@ -584,67 +638,96 @@ class BaseObject(with_metaclass(_BaseObjectMeta, StrConv, object)):
         for k in state:
             setattr(self, k, state[k])
 
-    if sys.version_info.major >= 3:
-        def __dir__(self):
-            return list(super(BaseObject, self).__dir__()) + list(self._fields.keys())
+    def __dir__(self):
+        return list(super(BaseObject, self).__dir__()) + list(self._fields.keys())
 
 
-class Currency(object):
+def _resolve_types(types):
+    actual_types = ()
+    for v in types:
+        if isinstance(v, str):
+            # the following is a (almost) copy/paste from
+            # https://stackoverflow.com/questions/11775460/lexical-cast-from-string-to-type
+            q = deque([object])
+            while q:
+                t = q.popleft()
+                if t.__name__ == v:
+                    actual_types += (t,)
+                else:
+                    try:
+                        # keep looking!
+                        q.extend(t.__subclasses__())
+                    except TypeError:
+                        # type.__subclasses__ needs an argument for
+                        # whatever reason.
+                        if t is type:
+                            continue
+                        else:
+                            raise
+        else:
+            actual_types += (v,)
+
+    return actual_types
+
+
+class Currency:
     CURRENCIES = OrderedDict([
-        (u'EUR', (u'€', u'EURO', u'EUROS')),
-        (u'CHF', (u'CHF',)),
-        (u'USD', (u'$', u'$US')),
-        (u'GBP', (u'£',)),
-        (u'LBP', (u'ل.ل',)),
-        (u'AED', (u'AED',)),
-        (u'XOF', (u'XOF',)),
-        (u'RUB', (u'руб',)),
-        (u'SGD', (u'SGD',)),
-        (u'BRL', (u'R$',)),
-        (u'MXN', (u'$',)),
-        (u'JPY', (u'¥',)),
-        (u'TRY', (u'₺', u'TRY')),
-        (u'RON', (u'lei',)),
-        (u'COP', (u'$',)),
-        (u'NOK', (u'kr',)),
-        (u'CNY', (u'¥',)),
-        (u'RSD', (u'din',)),
-        (u'ZAR', (u'rand',)),
-        (u'MYR', (u'RM',)),
-        (u'HUF', (u'Ft',)),
-        (u'HKD', (u'HK$',)),
-        (u'TWD', (u'NT$',)),
-        (u'QAR', (u'QR',)),
-        (u'MAD', (u'MAD',)),
-        (u'ARS', (u'ARS',)),
-        (u'AUD', (u'AUD',)),
-        (u'CAD', (u'CAD',)),
-        (u'NZD', (u'NZD',)),
-        (u'BHD', (u'BHD',)),
-        (u'SEK', (u'SEK',)),
-        (u'DKK', (u'DKK',)),
-        (u'LUF', (u'LUF',)),
-        (u'KZT', (u'KZT',)),
-        (u'PLN', (u'PLN',)),
-        (u'ILS', (u'ILS',)),
-        (u'THB', (u'THB',)),
-        (u'INR', (u'₹', u'INR')),
-        (u'PEN', (u'S/',)),
-        (u'IDR', (u'Rp',)),
-        (u'KWD', (u'KD',)),
-        (u'KRW', (u'₩',)),
-        (u'CZK', (u'Kč',)),
-        (u'EGP', (u'E£',)),
-        (u'ISK', (u'Íkr', u'kr')),
-        (u'XPF', (u'XPF',)),
-        (u'SAR', (u'SAR',)),
+        ('EUR', ('€', 'EURO', 'EUROS')),
+        ('CHF', ('CHF',)),
+        ('USD', ('$', '$US')),
+        ('GBP', ('£',)),
+        ('LBP', ('ل.ل',)),
+        ('AED', ('AED',)),
+        ('XOF', ('XOF',)),
+        ('RUB', ('руб',)),
+        ('SGD', ('SGD',)),
+        ('BRL', ('R$',)),
+        ('MXN', ('$',)),
+        ('JPY', ('¥',)),
+        ('TRY', ('₺', 'TRY')),
+        ('RON', ('lei',)),
+        ('COP', ('$',)),
+        ('NOK', ('kr',)),
+        ('CNY', ('¥',)),
+        ('RSD', ('din',)),
+        ('ZAR', ('rand',)),
+        ('MYR', ('RM',)),
+        ('HUF', ('Ft',)),
+        ('HKD', ('HK$',)),
+        ('TWD', ('NT$',)),
+        ('QAR', ('QR',)),
+        ('MAD', ('MAD',)),
+        ('ARS', ('ARS',)),
+        ('AUD', ('AUD',)),
+        ('CAD', ('CAD',)),
+        ('NZD', ('NZD',)),
+        ('BHD', ('BHD',)),
+        ('SEK', ('SEK',)),
+        ('DKK', ('DKK',)),
+        ('LUF', ('LUF',)),
+        ('KZT', ('KZT',)),
+        ('PLN', ('PLN',)),
+        ('ILS', ('ILS',)),
+        ('THB', ('THB',)),
+        ('INR', ('₹', 'INR')),
+        ('PEN', ('S/',)),
+        ('IDR', ('Rp',)),
+        ('KWD', ('KD',)),
+        ('KRW', ('₩',)),
+        ('CZK', ('Kč',)),
+        ('EGP', ('E£',)),
+        ('ISK', ('Íkr', 'kr')),
+        ('XPF', ('XPF',)),
+        ('SAR', ('SAR',)),
+        ('BGN', ('лв',)),
+        ('HRK', ('HRK',  'kn')),
     ])
 
     EXTRACTOR = re.compile(r'[()\d\s,\.\-]', re.UNICODE)
 
     @classmethod
     def get_currency(klass, text):
-        u"""
+        """
         >>> Currency.get_currency(u'42')
         None
         >>> Currency.get_currency(u'42 €')
@@ -673,7 +756,7 @@ class Currency(object):
 
     @classmethod
     def currency2txt(klass, currency):
-        _currency = klass.CURRENCIES.get(currency, (u'',))
+        _currency = klass.CURRENCIES.get(currency, ('',))
         return _currency[0]
 
 
