@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2012 Romain Bignon, Christophe Benz
 #
 # This file is part of woob.
@@ -18,9 +16,6 @@
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
 
-from __future__ import print_function
-
-import codecs
 import logging
 import optparse
 from optparse import OptionGroup, OptionParser
@@ -35,8 +30,7 @@ from woob.core.backendscfg import BackendsConfig
 from woob.tools.config.iconfig import ConfigError
 from woob.exceptions import FormFieldConversionWarning
 from woob.tools.log import createColoredFormatter, getLogger, DEBUG_FILTERS, settings as log_settings
-from woob.tools.misc import to_unicode, guess_encoding
-from woob.tools.compat import unicode
+from woob.tools.misc import guess_encoding
 
 from .results import ResultsConditionError
 
@@ -47,7 +41,7 @@ class MoreResultsAvailable(Exception):
     pass
 
 
-class ApplicationStorage(object):
+class ApplicationStorage:
     def __init__(self, name, storage):
         self.name = name
         self.storage = storage
@@ -75,7 +69,7 @@ class ApplicationStorage(object):
             return self.storage.save('applications', self.name)
 
 
-class Application(object):
+class Application:
     """
     Base application.
 
@@ -156,7 +150,7 @@ class Application(object):
     def __init__(self, option_parser=None):
         super(Application, self).__init__()
         self.encoding = self.guess_encoding()
-        self.logger = getLogger(self.APPNAME)
+        self.logger = getLogger("woob.applications.%s" % self.APPNAME)
         self.woob = self.create_woob()
         if self.CONFDIR is None:
             self.CONFDIR = self.woob.workdir
@@ -176,8 +170,12 @@ class Application(object):
             self._parser.add_option_group(app_options)
         self._parser.add_option('-b', '--backends', help='what backend(s) to enable (comma separated)')
         self._parser.add_option('-e', '--exclude-backends', help='what backend(s) to exclude (comma separated)')
+
         self._parser.add_option('-I', '--insecure', action='store_true', help='do not validate SSL')
         self._parser.add_option('--nss', action='store_true', help='Use NSS instead of OpenSSL')
+        self._parser.add_option('--force-ipv4', action='store_const', help='Force IPv4', const=4, dest="ipversion")
+        self._parser.add_option('--force-ipv6', action='store_const', help='Force IPv6', const=6, dest="ipversion")
+
         logging_options = OptionGroup(self._parser, 'Logging Options')
         logging_options.add_option('-d', '--debug', action='count', help='display debug messages. Set up it twice to more verbosity', default=0)
         logging_options.add_option('-q', '--quiet', action='store_true', help='display only error messages')
@@ -298,7 +296,7 @@ class Application(object):
             exclude = self.options.exclude_backends.split(',')
         loaded = self.woob.load_backends(caps, names, exclude=exclude, *args, **kwargs)
         if not loaded:
-            logging.info(u'No backend loaded')
+            logging.info('No backend loaded')
         return loaded
 
     def _get_optparse_version(self):
@@ -348,7 +346,7 @@ class Application(object):
         else:
             res = getattr(backend, function)(*args, **kwargs)
 
-        if hasattr(res, '__iter__') and not isinstance(res, (bytes, unicode)):
+        if hasattr(res, '__iter__') and not isinstance(res, (bytes, str)):
             return self._do_complete_iter(backend, count, selected_fields, res)
         else:
             return self._do_complete_obj(backend, selected_fields, res)
@@ -364,7 +362,7 @@ class Application(object):
         if isinstance(error, MoreResultsAvailable):
             return False
 
-        print(u'Error(%s): %s' % (backend.name, error), file=self.stderr)
+        print('Error(%s): %s' % (backend.name, error), file=self.stderr)
         if logging.root.level <= logging.DEBUG:
             print(backtrace, file=self.stderr)
         else:
@@ -379,7 +377,7 @@ class Application(object):
         :param errors: Object containing errors from backends
         :type errors: :class:`woob.core.bcall.CallErrors`
         :param debugmsg: Default message asking to enable the debug mode
-        :type debugmsg: :class:`basestring`
+        :type debugmsg: :class:`str`
         :param ignore: Exceptions to ignore
         :type ignore: tuple[:class:`Exception`]
         """
@@ -431,6 +429,8 @@ class Application(object):
             log_settings['ssl_insecure'] = True
         if self.options.nss:
             self.setup_nss()
+        if self.options.ipversion:
+            self.setup_ipversion()
 
         # this only matters to developers
         if not self.options.debug and not self.options.save_responses:
@@ -468,7 +468,8 @@ class Application(object):
         format = '%(asctime)s:%(levelname)s:%(name)s:' + cls.VERSION +\
                  ':%(filename)s:%(lineno)d:%(funcName)s %(message)s'
         handler = logging.StreamHandler(cls.stderr)
-        handler.setFormatter(createColoredFormatter(cls.stderr, format))
+        if os.getenv("NO_COLOR") is None:
+            handler.setFormatter(createColoredFormatter(cls.stderr, format))
         return handler
 
     @classmethod
@@ -490,6 +491,20 @@ class Application(object):
         init_nss(path)
         inject_in_urllib3()
 
+    def setup_ipversion(self):
+        import socket
+        import requests.packages.urllib3.util.connection
+        import urllib3.util.connection
+
+        for module in (requests.packages.urllib3.util.connection, urllib3.util.connection):
+            family = socket.AF_INET
+            if self.options.ipversion == 6 and module.HAS_IPV6:
+                family = socket.AF_INET6  # force ipv6 only if it is available
+
+            # unfortunately no other way than monkey-patching
+            # XXX though probably overkill, we need that to each `module`
+            module.allowed_gai_family = lambda family=family: family
+
     def create_logging_file_handler(self, filename):
         try:
             stream = open(os.path.expanduser(filename), 'w')
@@ -522,15 +537,8 @@ class Application(object):
 
         cls.setup_logging(logging.INFO, [cls.create_default_logger()])
 
-        if sys.version_info.major == 2:
-            encoding = sys.stdout.encoding
-            if encoding is None:
-                encoding = guess_encoding(sys.stdout)
-                cls.stdout = sys.stdout = codecs.getwriter(encoding)(sys.stdout)
-                # can't do the same with stdin, codecs.getreader buffers too much to be usable in a REPL
-
         if args is None:
-            args = [(cls.stdin.encoding and isinstance(arg, bytes) and arg.decode(cls.stdin.encoding) or to_unicode(arg)) for arg in sys.argv]
+            args = list(sys.argv)
 
         try:
             app = cls()
