@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright(C) 2010-2012  Christophe Benz, Romain Bignon
 #
 # This file is part of woob.
@@ -18,8 +16,6 @@
 # along with woob. If not, see <http://www.gnu.org/licenses/>.
 
 
-from __future__ import print_function
-
 from collections import OrderedDict
 from copy import copy
 import getpass
@@ -40,7 +36,6 @@ from woob.exceptions import BrowserUnavailable, BrowserIncorrectPassword, Browse
                               ActionNeeded, CaptchaQuestion, NeedInteractiveFor2FA
 from woob.tools.value import Value, ValueBool, ValueFloat, ValueInt, ValueBackendPassword
 from woob.tools.misc import to_unicode
-from woob.tools.compat import unicode, long
 
 from .base import Application, MoreResultsAvailable
 
@@ -89,6 +84,7 @@ class ConsoleApplication(Application):
     # shell escape strings
     if sys.platform == 'win32' \
             or not sys.stdout.isatty() \
+            or os.getenv('NO_COLOR') is not None \
             or os.getenv('ANSI_COLORS_DISABLED') is not None:
         #workaround to disable bold
         BOLD   = ''
@@ -131,7 +127,7 @@ class ConsoleApplication(Application):
         ret = super(ConsoleApplication, self).load_backends(*args, **kwargs)
 
         for err in errors:
-            print(u'Error(%s): %s' % (err.backend_name, err), file=self.stderr)
+            print('Error(%s): %s' % (err.backend_name, err), file=self.stderr)
             if self.ask('Do you want to reconfigure this backend?', default=True):
                 self.edit_backend(err.backend_name)
                 self.load_backends(names=[err.backend_name])
@@ -176,7 +172,7 @@ class ConsoleApplication(Application):
                                                        info.description))
             print('%sa) --all--%s               install all backends' % (self.BOLD, self.NC))
             print('%sq)%s --stop--\n' % (self.BOLD, self.NC))
-            r = self.ask('Select a backend to create (q to stop)', regexp='^(\d+|q|a)$')
+            r = self.ask('Select a backend to create (q to stop)', regexp=r'^(\d+|q|a)$')
 
             if str(r).isdigit():
                 i = int(r) - 1
@@ -296,7 +292,7 @@ class ConsoleApplication(Application):
             try:
                 backend.klass.register_account(account)
             except AccountRegisterError as e:
-                print(u'%s' % e)
+                print('%s' % e)
                 if self.ask('Do you want to try again?', default=True):
                     continue
                 else:
@@ -366,7 +362,7 @@ class ConsoleApplication(Application):
             if key not in params or edit:
                 params[key] = self.ask(value, default=params[key] if (key in params) else value.default)
             else:
-                print(u'[%s] %s: %s' % (key, value.description, '(masked)' if value.masked else to_unicode(params[key])))
+                print('[%s] %s: %s' % (key, value.description, '(masked)' if value.masked else to_unicode(params[key])))
         if asked_config:
             print('-------------------------%s' % ('-' * len(module.name)))
 
@@ -410,7 +406,7 @@ class ConsoleApplication(Application):
         if isinstance(question, Value):
             v = copy(question)
             if default is not None:
-                v.default = to_unicode(default) if isinstance(default, str) else default
+                v.default = default
             if masked is not None:
                 v.masked = masked
             if regexp is not None:
@@ -424,7 +420,7 @@ class ConsoleApplication(Application):
                 klass = ValueBool
             elif isinstance(default, float):
                 klass = ValueFloat
-            elif isinstance(default, (int,long)):
+            elif isinstance(default, int):
                 klass = ValueInt
             else:
                 klass = Value
@@ -433,9 +429,9 @@ class ConsoleApplication(Application):
 
         question = v.label
         if v.description and v.description != v.label:
-            question = u'%s: %s' % (question, v.description)
+            question = '%s: %s' % (question, v.description)
         if v.id:
-            question = u'[%s] %s' % (v.id, question)
+            question = '[%s] %s' % (v.id, question)
 
         if isinstance(v, ValueBackendPassword):
             print(question + ':')
@@ -459,7 +455,7 @@ class ConsoleApplication(Application):
                 while True:
                     cmd = self.ask('')
                     try:
-                        check_output(cmd, shell=True)
+                        check_output(cmd, shell=True)  # nosec: this is intended
                     except subprocess.CalledProcessError as e:
                         print('%s' % e)
                     else:
@@ -467,7 +463,7 @@ class ConsoleApplication(Application):
 
         aliases = {}
         if isinstance(v, ValueBool):
-            question = u'%s (%s/%s)' % (question, 'Y' if v.default else 'y', 'n' if v.default else 'N')
+            question = '%s (%s/%s)' % (question, 'Y' if v.default else 'y', 'n' if v.default else 'N')
         elif v.choices:
             if v.tiny is None:
                 v.tiny = True
@@ -477,7 +473,7 @@ class ConsoleApplication(Application):
                         break
 
             if v.tiny:
-                question = u'%s (%s)' % (question, '/'.join((s.upper() if s == v.default else s)
+                question = '%s (%s)' % (question, '/'.join((s.upper() if s == v.default else s)
                                                             for s in v.choices))
                 for s in v.choices:
                     if s == v.default:
@@ -488,20 +484,17 @@ class ConsoleApplication(Application):
                 for n, (key, value) in enumerate(v.choices.items()):
                     print('     %s%2d)%s %s' % (self.BOLD, n + 1, self.NC, value))
                     aliases[str(n + 1)] = key
-                question = u'%s (choose in list)' % question
+                question = '%s (choose in list)' % question
         if v.masked:
-            question = u'%s (hidden input)' % question
+            question = '%s (hidden input)' % question
 
         if not isinstance(v, ValueBool) and not v.tiny and v.default not in (None, ''):
-            question = u'%s [%s]' % (question, '*******' if v.masked else v.default)
+            question = '%s [%s]' % (question, '*******' if v.masked else v.default)
 
         question += ': '
 
         while True:
             if v.masked:
-                if sys.version_info.major < 3 and isinstance(question, unicode):
-                    question = question.encode(self.encoding)
-
                 line = getpass.getpass(question)
                 if sys.platform != 'win32':
                     if isinstance(line, bytes): # only for python2
@@ -509,7 +502,7 @@ class ConsoleApplication(Application):
             else:
                 self.stdout.write(question)
                 self.stdout.flush()
-                line = self._readline()
+                line = self.stdin.readline()
                 if len(line) == 0:
                     raise EOFError()
                 else:
@@ -524,7 +517,7 @@ class ConsoleApplication(Application):
             try:
                 v.set(line)
             except ValueError as e:
-                print(u'Error: %s' % e, file=self.stderr)
+                print('Error: %s' % e, file=self.stderr)
             else:
                 break
 
@@ -534,18 +527,6 @@ class ConsoleApplication(Application):
     def print(self, txt):
         print(txt)
 
-    def _readall(self):
-        if sys.version_info.major == 2:
-            return self.stdin.read().decode(self.encoding)
-        else:
-            return self.stdin.read()
-
-    def _readline(self):
-        if sys.version_info.major == 2:
-            return self.stdin.readline().decode(self.encoding)
-        else:
-            return self.stdin.readline()
-
     def acquire_input(self, content=None, editor_params=None):
         editor = os.getenv('EDITOR', 'vi')
         if self.stdin.isatty() and editor:
@@ -553,7 +534,7 @@ class ConsoleApplication(Application):
             with NamedTemporaryFile() as f:
                 filename = f.name
                 if content is not None:
-                    if isinstance(content, unicode):
+                    if isinstance(content, str):
                         content = content
                     f.write(content)
                     f.flush()
@@ -569,7 +550,7 @@ class ConsoleApplication(Application):
             if self.stdin.isatty():
                 print('Reading content from stdin... Type ctrl-D '
                           'from an empty line to stop.')
-            text = self._readall()
+            text = self.stdin.read()
         return to_unicode(text)
 
     def bcall_error_handler(self, backend, error, backtrace):
@@ -584,9 +565,9 @@ class ConsoleApplication(Application):
                 if v:
                     backend.config[field.id].set(v)
         elif isinstance(error, CaptchaQuestion):
-            print(u'Warning(%s): Captcha has been found on login page' % backend.name, file=self.stderr)
+            print('Warning(%s): Captcha has been found on login page' % backend.name, file=self.stderr)
         elif isinstance(error, BrowserIncorrectPassword):
-            msg = unicode(error)
+            msg = str(error)
             if not msg:
                 msg = 'invalid login/password.'
             print('Error(%s): %s' % (backend.name, msg), file=self.stderr)
@@ -595,32 +576,32 @@ class ConsoleApplication(Application):
                 self.edit_backend(backend.name)
                 self.load_backends(names=[backend.name])
         elif isinstance(error, BrowserSSLError):
-            print(u'FATAL(%s): ' % backend.name + self.BOLD + '/!\ SERVER CERTIFICATE IS INVALID /!\\' + self.NC, file=self.stderr)
+            print('FATAL(%s): ' % backend.name + self.BOLD + r'/!\ SERVER CERTIFICATE IS INVALID /!\\' + self.NC, file=self.stderr)
         elif isinstance(error, BrowserHTTPSDowngrade):
-            print(u'FATAL(%s): ' % backend.name + 'Downgrade from HTTPS to HTTP')
+            print('FATAL(%s): ' % backend.name + 'Downgrade from HTTPS to HTTP')
         elif isinstance(error, BrowserForbidden):
-            msg = unicode(error)
-            print(u'Error(%s): %s' % (backend.name, msg or 'Forbidden'), file=self.stderr)
+            msg = str(error)
+            print('Error(%s): %s' % (backend.name, msg or 'Forbidden'), file=self.stderr)
         elif isinstance(error, BrowserUnavailable):
-            msg = unicode(error)
-            print(u'Error(%s): %s' % (backend.name, msg or 'Website is unavailable.'), file=self.stderr)
+            msg = str(error)
+            print('Error(%s): %s' % (backend.name, msg or 'Website is unavailable.'), file=self.stderr)
         elif isinstance(error, ActionNeeded):
-            msg = unicode(error)
-            print(u'Error(%s): Action needed on website: %s' % (backend.name, msg), file=self.stderr)
+            msg = str(error)
+            print('Error(%s): Action needed on website: %s' % (backend.name, msg), file=self.stderr)
         elif isinstance(error, NotImplementedError):
-            print(u'Error(%s): this feature is not supported yet by this backend.' % backend.name, file=self.stderr)
-            print(u'      %s   To help the maintainer of this backend implement this feature,' % (' ' * len(backend.name)), file=self.stderr)
-            print(u'      %s   please contact us on the project mailing list' % (' ' * len(backend.name)), file=self.stderr)
+            print('Error(%s): this feature is not supported yet by this backend.' % backend.name, file=self.stderr)
+            print('      %s   To help the maintainer of this backend implement this feature,' % (' ' * len(backend.name)), file=self.stderr)
+            print('      %s   please contact us on the project mailing list' % (' ' * len(backend.name)), file=self.stderr)
         elif isinstance(error, NeedInteractiveFor2FA):
-            print(u'Error(%s): You have to run %s in interactive mode to perform a two-factor authentication' % (backend.name, self.APPNAME), file=self.stderr)
+            print('Error(%s): You have to run %s in interactive mode to perform a two-factor authentication' % (backend.name, self.APPNAME), file=self.stderr)
         elif isinstance(error, UserError):
-            print(u'Error(%s): %s' % (backend.name, to_unicode(error)), file=self.stderr)
+            print('Error(%s): %s' % (backend.name, error), file=self.stderr)
         elif isinstance(error, MoreResultsAvailable):
-            print(u'Hint: There are more results for backend %s' % (backend.name), file=self.stderr)
+            print('Hint: There are more results for backend %s' % (backend.name), file=self.stderr)
         elif isinstance(error, NoAccountsException):
-            print(u'Error(%s): %s' % (backend.name, to_unicode(error) or 'No account on this backend'), file=self.stderr)
+            print('Error(%s): %s' % (backend.name, str(error) or 'No account on this backend'), file=self.stderr)
         else:
-            print(u'Bug(%s): %s' % (backend.name, to_unicode(error)), file=self.stderr)
+            print('Bug(%s): %s' % (backend.name, error), file=self.stderr)
 
             minfo = self.woob.repositories.get_module_info(backend.NAME)
             if minfo and not minfo.is_local():
